import torchelie.nn
import torchelie.utils
import torchelie.loss
import torchelie.transforms
import torchelie.models
import torchelie.callbacks
import torchelie.datasets
import torchelie.data_learning
import torchelie.optim
import torchelie.distributions
import torchelie.lr_scheduler
import torchelie.hyper
import torchelie.recipes

__version__ = '0.1.0'



from typing import Tuple
import copy
import os
import torch
import torchelie as tch
import torchelie.utils as tu
from torchelie.recipes.gan import GANRecipe
from torchelie.transforms import MultiBranch
import torchvision.transforms as TF
import torchelie.loss.gan.standard as gan_loss
from torchelie.loss.gan.penalty import zero_gp
from torchelie.datasets import UnlabeledImages, Pix2PixDataset
from torchelie.datasets import SideBySideImagePairsDataset
from torchelie.models import *
import torch.nn as nn


def get_dataset(dataset_specs: Tuple[str, str], img_size: int):
    ty, path = dataset_specs
    if ty == 'pix2pix':
        return Pix2PixDataset('~/.torch',
                              path,
                              split='train',
                              download=True,
                              transform=TF.Compose([
                                  TF.Resize(img_size),
                                  TF.RandomResizedCrop(img_size,
                                                       scale=(0.9, 1)),
                                  TF.RandomHorizontalFlip(),
                              ]))
    if ty == 'colorize':
        return UnlabeledImages(
            path,
            TF.Compose([
                TF.Resize(img_size),
                TF.RandomCrop(img_size),
                TF.RandomHorizontalFlip(),
                MultiBranch([
                    TF.Compose([
                        TF.Grayscale(3),
                        TF.ToTensor(),
                    ]),
                    TF.ToTensor(),
                ])
            ]))
    if ty == 'inpainting':
        return UnlabeledImages(
            path,
            TF.Compose([
                TF.Resize(img_size),
                TF.RandomCrop(img_size),
                TF.RandomHorizontalFlip(),
                TF.ToTensor(),
                MultiBranch([
                    TF.Compose([
                        TF.RandomErasing(p=1, value=(1., 1., 0)),
                    ]),
                    TF.Compose([])
                ])
            ]))
    if ty == 'edges':
        return UnlabeledImages(
            path,
            TF.Compose([
                TF.Resize(img_size),
                TF.CenterCrop(img_size),
                TF.RandomHorizontalFlip(),
                MultiBranch([
                    TF.Compose([
                        tch.transforms.Canny(),
                        TF.Grayscale(3),
                        TF.ToTensor(),
                    ]),
                    TF.Compose([
                        TF.ToTensor(),
                    ])
                ])
            ]))
    if ty == 'pairs':
        return SideBySideImagePairsDataset(
            path,
            TF.Compose([
                TF.Resize(img_size),
                TF.CenterCrop(img_size),
                TF.RandomHorizontalFlip(),
                TF.ToTensor(),
            ]))
    assert False, "dataset's type not understood"


@tu.experimental
def train(rank, world_size):
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument('--dataset', required=True, type=lambda x: x.split(':'))
    parser.add_argument('--r0-gamma', type=float)
    parser.add_argument('--D-type', choices=['patch', 'unet'], default='patch')
    parser.add_argument('--l1-gain', default=0, type=float)
    parser.add_argument('--batch-size', default=4, type=int)
    parser.add_argument('--from-ckpt')
    opts = parser.parse_args()

    G = pix2pix_256().to_instance_norm().to_equal_lr()
    G_polyak = copy.deepcopy(G)
    if opts.D_type == 'patch':
        D = residual_patch286()
        D.set_input_specs(6)
        D.to_equal_lr()
        r0_gamma = 0.1
    else:
        D = UNet([32, 64, 128, 256, 512], 1)
        D.set_decoder_num_layers(1)
        D.set_encoder_num_layers(1)
        D.set_input_specs(6)
        D.to_bilinear_sampling()
        D.leaky()
        tnn.utils.net_to_equal_lr(D, leak=0.2)
        r0_gamma = 0.00001
    r0_gamma = opts.r0_gamma or r0_gamma

    if rank == 0:
        print(G)
        print(D)

    G = torch.nn.parallel.DistributedDataParallel(G.to(rank), [rank], rank)
    D = torch.nn.parallel.DistributedDataParallel(D.to(rank), [rank], rank)

    ds = get_dataset(opts.dataset, 256)
    ds = torch.utils.data.DataLoader(ds,
                                     opts.batch_size,
                                     num_workers=4,
                                     shuffle=True,
                                     pin_memory=True,
                                     drop_last=True)

    def G_fun(batch) -> dict:
        x, y = batch
        G.train()
        D.train()
        out = G(x * 2 - 1)
        with D.no_sync():
            loss = gan_loss.generated(D(torch.cat([out, x], dim=1) * 2 - 1))
        loss += opts.l1_gain * F.l1_loss(out, y)
        loss.backward()
        return {'G_loss': loss.item()}

    class GradientPenalty:

        @tu.experimental
        def __init__(self, gamma):
            self.gamma = gamma
            self.iters = 0
            self.last_norm = float('nan')

        def __call__(self, model, real, fake):
            if self.iters < 100 or self.iters % 4 == 0:
                real = real.detach()
                fake = fake.detach()
                gp, g_norm = zero_gp(model, real, fake)
                # Sync the gradient on the next backward
                if torch.any(torch.isnan(gp)):
                    gp.detach_()
                else:
                    (4 * self.gamma * gp).backward()
                self.last_norm = g_norm
            self.iters += 1
            return self.last_norm

    gradient_penalty = GradientPenalty(r0_gamma)

    def D_fun(batch) -> dict:
        G.train()
        D.train()
        x, y = batch
        with G.no_sync():
            with torch.no_grad():
                out = G(x * 2 - 1)
        fake = torch.cat([out, x], dim=1) * 2 - 1
        real = torch.cat([y, x], dim=1) * 2 - 1
        with D.no_sync():
            prob_fake = D(fake)
            fake_correct = prob_fake.detach().lt(0).int().eq(1).sum()
            fake_loss = gan_loss.fake(prob_fake)
            fake_loss.backward()

        with D.no_sync():
            g_norm = gradient_penalty(D, real, fake)

        prob_real = D(real)
        real_correct = prob_real.detach().gt(0).int().eq(1).sum()
        real_loss = gan_loss.real(prob_real)
        real_loss.backward()

        return {
            'out': out.detach(),
            'fake_loss': fake_loss.item(),
            'prob_fake': torch.sigmoid(prob_fake).mean().item(),
            'fake_heatmap': torch.sigmoid(prob_fake.detach()),
            'prob_real': torch.sigmoid(prob_real).mean().item(),
            'real_loss': real_loss.item(),
            'g_norm': g_norm,
            'D-correct': (fake_correct + real_correct) / (2 * prob_fake.numel())
        }

    def test_fun(batch):
        x, _ = batch
        G_polyak.train()
        out = G_polyak(x * 2 - 1)
        return {'out': out.detach()}

    tag = f'pix2pix_{opts.dataset[0]}:{os.path.basename(opts.dataset[1])}'
    recipe = GANRecipe(G,
                       D,
                       G_fun,
                       D_fun,
                       test_fun,
                       ds,
                       checkpoint=tag if rank == 0 else None,
                       visdom_env=tag if rank == 0 else None)

    recipe.register('G_polyak', G_polyak)

    recipe.callbacks.add_callbacks([
        tch.callbacks.Optimizer(
            tch.optim.RAdamW(D.parameters(),
                             lr=2e-3,
                             betas=(0., 0.99),
                             weight_decay=0)),
        tch.callbacks.Log('out', 'out'),
        tch.callbacks.Log('batch.0', 'x'),
        tch.callbacks.Log('batch.1', 'y'),
        tch.callbacks.Log('fake_heatmap', 'fake_heatmap'),
        tch.callbacks.WindowedMetricAvg('fake_loss', 'fake_loss'),
        tch.callbacks.WindowedMetricAvg('real_loss', 'real_loss'),
        tch.callbacks.WindowedMetricAvg('prob_fake', 'prob_fake'),
        tch.callbacks.WindowedMetricAvg('prob_real', 'prob_real'),
        tch.callbacks.WindowedMetricAvg('D-correct', 'D-correct'),
        tch.callbacks.Log('g_norm', 'g_norm')
    ])
    recipe.G_loop.callbacks.add_callbacks([
        tch.callbacks.Optimizer(
            tch.optim.RAdamW(G.parameters(),
                             lr=2e-3,
                             betas=(0., 0.99),
                             weight_decay=0)),
        tch.callbacks.Polyak(G.module, G_polyak),
    ])
    recipe.test_loop.callbacks.add_callbacks([
        tch.callbacks.Log('out', 'polyak_out'),
    ])
    recipe.to(rank)
    if opts.from_ckpt is not None:
        recipe.load_state_dict(torch.load(opts.from_ckpt, map_location='cpu'))
    recipe.run(200)


if __name__ == '__main__':
    tu.parallel_run(train)



import torch.nn as nn
from typing import Union, Tuple, List
from torchelie.utils import experimental


def tup(x):
    if isinstance(x, (tuple, list)):
        return list(x)
    return [x]


ArgNames = Union[str, List[str]]
NamedModule = Tuple[str, nn.Module]


class ModuleGraph(nn.Sequential):
    """
    Allows description of networks as computation graphs. The graph is
    constructed by labelling inputs and outputs of each node. Each node will be
    ran in declaration order, fetching its input values from a pool of named
    values populated from previous node's output values and keyword arguments
    in forward.

    Simple example:

    >>> m = tnn.ModuleGraph(outputs='y')
    >>> m.add_operation(
            inputs=['x'],
            operation=nn.Linear(10, 20),
            name='linear',
            outputs=['y'])
    >>> m(x=torch.randn(1, 10))
    <a bunch of numbers>

    Multiple inputs example:

    If a layer takes more than 1 input, labels can be a tuple or a list of
    labels instead. The same applies if a module returns more than 1 output
    values.

    >>> m = tnn.ModuleGraph(outputs=['x1', 'y'])
    >>> m.add_operation(
            inputs=['x0'],
            operation=nn.Linear(10, 20)
            name='linear',
            outputs=['x1'])
    >>> m.add_operation(
            inputs=['x1', 'z'],
            operation=nn.AdaIN2d(20, 3)
            name='adain',
            outputs=['y'])
    >>> m(x0=torch.randn(1, 10), z=torch.randn(1, 3))['y']
    <a bunch of numbers>
    """
    def __init__(self, outputs: Union[str, List[str]]) -> None:
        super().__init__()
        self.ins: List[List[str]] = []
        self.outs: List[List[str]] = []

        self.outputs = outputs

    def add_operation(self, inputs: List[str], outputs: List[str], name: str,
                      operation: nn.Module) -> 'ModuleGraph':
        self.ins.append(inputs)
        self.outs.append(outputs)
        self.add_module(name, operation)
        return self

    def forward(self, **args):
        variables = dict(args)

        for i_names, f, o_names in zip(self.ins, self._modules.values(),
                                       self.outs):
            ins = [variables[k] for k in i_names]
            outs = tup(f(*ins))
            for o, k in zip(outs, o_names):
                variables[k] = o

        if isinstance(self.outputs, str):
            return variables[self.outputs]
        return {k: variables[k] for k in self.outputs}

    @experimental
    def to_dot(self) -> str:
        txt = ''
        for i_names, f_nm, o_names in zip(self.ins, self._modules.keys(),
                                          self.outs):
            for k in i_names:
                txt += f'{k} -> {f_nm};\n'
            for k in o_names:
                txt += f'{f_nm} -> {k};\n'
            txt += f'{f_nm} [shape=square];\n'
        return txt



import torch
import torchelie.utils as tu
import torch.nn as nn
import torchelie.nn as tnn
from torchelie.models import ClassificationHead
from typing import Optional
from collections import OrderedDict

Block = tnn.PreactResBlockBottleneck


class UBlock(nn.Module):
    inner: Optional[nn.Module]
    skip: Optional[nn.Module]
    encode: nn.Module

    @tu.experimental
    def __init__(self,
                 ch: int,
                 inner: Optional[nn.Module],
                 with_skip: bool = True) -> None:
        super(UBlock, self).__init__()
        self.inner = inner
        if with_skip and inner is not None:
            self.skip = Block(ch, ch)
        else:
            self.skip = None
        self.encode = tnn.CondSeq(nn.MaxPool2d(3, 1, 1),
                                  nn.UpsamplingBilinear2d(scale_factor=0.5),
                                  Block(ch, ch))
        self.decode = tnn.CondSeq(Block(ch, ch),
                                  nn.UpsamplingBilinear2d(scale_factor=2))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        e = self.encode(x)
        if self.inner is not None:
            e2 = self.inner(e)
        else:
            e2 = e

        if self.skip is not None:
            e2 += self.skip(e)

        return self.decode(e2)


class UBlock1(nn.Module):
    @tu.experimental
    def __init__(self, ch):
        super(UBlock1, self).__init__()
        self.inner = tnn.CondSeq(nn.MaxPool2d(3, 1, 1),
                                 nn.UpsamplingBilinear2d(scale_factor=0.5),
                                 Block(ch, ch),
                                 nn.UpsamplingBilinear2d(scale_factor=2))

    def forward(self, x):
        return self.inner(x)


class AttentionBlock(nn.Module):
    mask: Optional[tnn.CondSeq]

    def __init__(self,
                 ch: int,
                 n_down: int,
                 n_trunk: int = 2,
                 n_post: int = 1,
                 n_pre: int = 1,
                 n_att_conv: int = 2,
                 with_skips: bool = True) -> None:
        super(AttentionBlock, self).__init__()
        self.pre = tnn.CondSeq(*[Block(ch, ch) for _ in range(n_pre)])
        self.post = tnn.CondSeq(*[Block(ch, ch) for _ in range(n_post)])
        self.trunk = tnn.CondSeq(*[Block(ch, ch) for _ in range(n_trunk)])

        soft: nn.Module = UBlock1(ch)
        for _ in range(n_down - 1):
            soft = UBlock(ch, soft, with_skip=with_skips)

        if n_down >= 0:
            conv1 = [soft]
            for i in range(n_att_conv):
                conv1 += [
                    nn.BatchNorm2d(ch),
                    nn.ReLU(True),
                    tu.kaiming(tnn.Conv1x1(ch, ch, bias=(i != n_att_conv - 1)))
                ]
            conv1.append(nn.Sigmoid())

            self.mask = tnn.CondSeq(*conv1)
        else:
            self.mask = None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.pre(x)
        t = self.trunk(x)
        if self.mask is not None:
            t = t * (self.mask(x) + 1)
        return self.post(t)


class Attention56Bone(tnn.CondSeq):
    """
    Attention56 bone

    Args:
        in_ch (int): number of channels in the images
    """
    @tu.experimental
    def __init__(self, num_classes: int) -> None:
        super(Attention56Bone, self).__init__(
            OrderedDict([
                ('head',
                 tnn.CondSeq(tu.kaiming(tnn.Conv2d(3, 64, 7, stride=2)),
                             nn.ReLU(True), nn.MaxPool2d(3, 2, 1))),
                ('pre1', Block(64, 256)), ('attn1', AttentionBlock(256, 3)),
                ('pre2', Block(256, 512, stride=2)),
                ('attn2', AttentionBlock(512, 2)),
                ('pre3', Block(512, 1024, stride=2)),
                ('attn3', AttentionBlock(1024, 1)),
                ('pre4',
                 tnn.CondSeq(
                     Block(1024, 2048, stride=2),
                     Block(2048, 2048),
                     Block(2048, 2048),
                 )),
                ('classifier', ClassificationHead(2048, num_classes))
            ]))


@tu.experimental
def attention56(num_classes):
    """
    Build a attention56 network

    Args:
        num_classes (int): number of classes
        in_ch (int): number of channels in the images
    """
    return Attention56Bone(num_classes)



import random
import math
import torch
import torch.nn as nn
import torchelie.utils as tu
import torchelie.nn as tnn
from typing import List, Tuple, Union, overload
from typing_extensions import Literal
from .classifier import ClassificationHead
from .snres_discr import ResidualDiscriminator
from torchelie.transforms.differentiable import BinomialFilter2d


class LinearReLU(tnn.CondSeq):
    relu: nn.Module
    linear: nn.Module

    @tu.experimental
    def __init__(self, in_features: int, out_features) -> None:
        super().__init__()
        self.linear = tu.kaiming(nn.Linear(in_features, out_features))
        self.relu = nn.ReLU(True)

    def leaky(self) -> 'LinearReLU':
        if isinstance(self.relu, (nn.ReLU, nn.LeakyReLU)):
            self.relu = nn.LeakyReLU(0.2, self.relu.inplace)
        else:
            self.relu = nn.LeakyReLU(0.2, True)
        return self

    def to_equal_lr(self) -> 'LinearReLU':
        if isinstance(self.relu, nn.LeakyReLU):
            tu.kaiming(self.linear, dynamic=True, a=self.relu.negative_slope)
        else:
            tu.kaiming(self.linear, dynamic=True)
        return self

    def to_differential_lr(self, lr: float):
        tnn.utils.weight_scale(self.linear, name='bias', scale=lr)
        if isinstance(self.relu, nn.LeakyReLU):
            g = tu.kaiming_gain(self.linear, a=self.relu.negative_slope)
        else:
            g = tu.kaiming_gain(self.linear)
        tnn.utils.weight_scale(self.linear, scale=lr * g)
        assert isinstance(self.linear.weight_g, torch.Tensor)
        self.linear.weight_g.data.normal_(0, 1. / lr)
        return self


class MappingNetwork(tnn.CondSeq):
    def __init__(self, num_features: int, num_layers: int = 3) -> None:
        super().__init__()
        for i in range(num_layers):
            self.add_module(f'linear_{i}',
                            LinearReLU(num_features, num_features))

    def leaky(self) -> 'MappingNetwork':
        for m in self:
            m.leaky()
        return self

    def to_equal_lr(self) -> 'MappingNetwork':
        for m in self:
            m.to_equal_lr()
        return self

    def to_differential_lr(self, lr: float) -> 'MappingNetwork':
        for m in self:
            m.to_differential_lr(lr)
        return self


class StyleGAN2Generator(nn.Module):
    """
    Generator from StyleGAN2

    Args:
        noise_size (int): size of the input gaussian noise
        ch_mul (float): width adaptor, multiplies the number of channels each
            layer (default: 1.)
        img_size (int): spatial size of the image to generate (default: 128)
        max_ch (int): maximum number of channels (default: 512)
    """
    w_avg: torch.Tensor

    @tu.experimental
    def __init__(self,
                 noise_size: int,
                 ch_mul: float = 1,
                 img_size: int = 128,
                 max_ch: int = 512):
        super().__init__()
        self.register_buffer('w_avg', torch.zeros(noise_size))

        self.encode = MappingNetwork(noise_size, 3).to_differential_lr(0.01)
        self.encode.leaky()
        res = 4
        render = tnn.ModuleGraph('out')

        while res <= img_size:
            ch = int(512 / (2**(math.log2(res) - 7)) * ch_mul)
            if res == 4:
                render.add_operation(inputs=['N'],
                                     outputs=['fmap_constxconst'],
                                     name='const',
                                     operation=tnn.Const(
                                         min(max_ch, ch), 4, 4))
                prev_res = 'const'
            else:
                prev_res = str(res // 2)

            block = tnn.StyleGAN2Block(min(max_ch, ch),
                                       min(max_ch, ch // 2),
                                       noise_size,
                                       upsample=res != 4,
                                       n_layers=1 if res == 4 else 2)
            render.add_operation(
                inputs=[
                    f'fmap_{prev_res}x{prev_res}', f'w_{res}x{res}',
                    f'rgb_{prev_res}x{prev_res}'
                ],
                operation=block,
                name=f'conv{res}x{res}',
                outputs=[f'rgb_{res}x{res}', f'fmap_{res}x{res}'])
            res *= 2

        render.add_operation(inputs=[f'rgb_{img_size}x{img_size}'],
                             operation=nn.Sigmoid(),
                             name='sigmoid',
                             outputs=['out'])

        self.render = render

    @overload
    def forward(self,
                z,
                mixing: bool,
                get_w: Literal[False]) -> torch.Tensor:
        ...

    @overload
    def forward(
            self,
            z,
            mixing: bool,
            get_w: Literal[True]) -> Tuple[torch.Tensor, torch.Tensor]:
        ...

    def forward(self, z, mixing: bool = True, get_w: bool = False):
        w = self.encode(z)

        if self.training:
            self.w_avg.copy_(tu.lerp(w.detach().mean(0), self.w_avg, 0.995))

        L = random.randint(0, len(self.render) - 1)
        mix_res = 2**(L + 2)
        ws = {'N': w.shape[0]}
        for i_res in range(len(self.render)):
            res = 2**(i_res + 2)
            if mixing and res == mix_res and random.uniform(0, 1) < 0.9:
                w = self.encode(torch.randn_like(z))

            if res <= 32 and not self.training:
                ws[f'w_{res}x{res}'] = 0.3 * self.w_avg + 0.7 * w
            else:
                ws[f'w_{res}x{res}'] = w

        if not get_w:
            return self.render(rgb_constxconst=None, **ws)
        else:
            return self.render(rgb_constxconst=None, **ws), w

    def w_to_dict(self, w):
        return {
            'N': w.shape[0],
            **{
                f'w_{2**(i+2)}x{2**(i+2)}': w
                for i in range(len(self.render) - 1)
            }
        }


@tu.experimental
def StyleGAN2Discriminator(input_sz,
                           max_ch: int = 512,
                           ch_mul: int = 1) -> ResidualDiscriminator:
    """
        Build the discriminator for StyleGAN2

        Args:
            input_sz (int): image size
            max_ch (int): maximum number of channels (default: 512)
            ch_mul (float): multiply the number of channels on each layer by this
                value (default, 1.)
    """
    import math
    res = input_sz
    ch = int(512 / (2**(math.log2(res) - 6)) * ch_mul)
    layers: List[Union[str, int]] = [min(max_ch, ch)]

    while res > 4:
        res = res // 2
        layers.append('D')
        layers.append(min(max_ch, ch * 2))
        ch *= 2

    net = ResidualDiscriminator(layers)

    def make_binomial(m):
        if isinstance(m, nn.AvgPool2d):
            return BinomialFilter2d(2)
        return m

    tnn.utils.edit_model(net, make_binomial)
    net.add_minibatch_stddev()
    assert isinstance(net.classifier, ClassificationHead)
    net.classifier.to_two_layers(min(ch, max_ch))
    net.classifier.set_pool_size(4)
    net.classifier.leaky()
    net.to_equal_lr()
    return net



import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import List, Optional


class Interpolate2d(nn.Module):
    """
    A wrapper around :func:`pytorch.nn.functional.interpolate`
    """
    def __init__(self,
                 mode: str,
                 size: Optional[List[int]] = None,
                 scale_factor: Optional[float] = None) -> None:
        super().__init__()
        self.size = size
        self.scale_factor = scale_factor
        self.mode = mode

    def forward(self,
                x: torch.Tensor,
                size: Optional[List[int]] = None) -> torch.Tensor:
        rsf = True if self.scale_factor is not None else None
        if not size:
            return F.interpolate(x,
                                 mode=self.mode,
                                 size=self.size,
                                 scale_factor=self.scale_factor,
                                 recompute_scale_factor=rsf,
                                 align_corners=False)
        else:
            return F.interpolate(x,
                                 mode=self.mode,
                                 size=size,
                                 recompute_scale_factor=rsf,
                                 align_corners=False)

    def extra_repr(self) -> str:
        return f'scale_factor={self.scale_factor} size={self.size}'


class InterpolateBilinear2d(Interpolate2d):
    """
    A wrapper around :func:`pytorch.nn.functional.interpolate` with bilinear
    mode.
    """
    def __init__(
        self,
        size: Optional[List[int]] = None,
        scale_factor: Optional[float] = None,
    ) -> None:
        super().__init__(size=size, scale_factor=scale_factor, mode='bilinear')



import random
import copy
import math
import torch
import torch.nn as nn
import torchelie as tch
import torchelie.utils as tu
import torchelie.callbacks as tcb
import torchelie.loss.gan.standard as gan_loss
from torchelie.optim import RAdamW
from torchelie.loss.gan.penalty import zero_gp
from torchelie.recipes.gan import GANRecipe
from torchelie.callbacks.avg import ExponentialAvg


@tu.experimental
class ADATF:
    p: float

    def __init__(self, target_loss: float, growth: float = 0.01):
        self.p = 0
        self.target_loss = target_loss
        self.growth = growth
        self.loss = ExponentialAvg(0.95)

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        if self.p == 0:
            return x
        p = self.p
        RND = (x.shape[0], 1, 1, 1)
        x = torch.where(
            torch.rand(RND, device=x.device) < p / 2,
            tch.transforms.differentiable.roll(x, random.randint(-16, 16),
                                               random.randint(-16, 16)), x)
        color = tch.transforms.differentiable.AllAtOnceColor(x.shape[0])
        color.brightness(0.5, p)
        color.contrast(0.7, p)
        x = color.apply(x)

        geom = tch.transforms.differentiable.AllAtOnceGeometric(x.shape[0])
        geom.translate(0.25, 0.25, p)
        geom.rotate(180, p)
        geom.scale(0.5, 0.5, p)
        geom.flip_x(0.5, p)
        geom.flip_y(0.5, p)
        x = geom.apply(x)
        return x

    def log_loss(self, loss: float) -> None:
        self.loss.log(loss)
        if self.loss.get() > self.target_loss:
            self.p -= self.growth
        else:
            self.p += self.growth
        self.p = max(0, min(self.p, 0.9))


class PPL:
    @tu.experimental
    def __init__(self, every=4):
        self.every = every
        self.ppl_goal = 0.
        self.iters = 0

    def compute_ppl(self, model, z):
        gen, w = model(z, get_w=True, mixing=False)
        B, C, H, W = gen.shape
        noise = torch.randn_like(gen) / math.sqrt(H * W)
        JwTy = torch.autograd.grad(outputs=torch.sum(gen * noise),
                                   inputs=w,
                                   create_graph=True,
                                   only_inputs=True)[0]
        JwTy_norm = JwTy.pow(2).sum(1).mul(
            1 / (2 * len(model.module.render))).sqrt()

        E_JwTy_norm = JwTy_norm.detach().mean().item()
        if self.ppl_goal is None:
            self.ppl_goal = E_JwTy_norm
        self.ppl_goal = 0.99 * self.ppl_goal + 0.01 * E_JwTy_norm

        return (JwTy_norm - self.ppl_goal).pow(2).mean() * 2

    def __call__(self, model, z):
        ppl = self.ppl_goal
        if self.iters % self.every == 0:
            ppl = self.compute_ppl(model, z)
            (self.every * ppl).backward()

        self.iters += 1
        return self.ppl_goal


class GradientPenalty:
    @tu.experimental
    def __init__(self, gamma):
        self.gamma = gamma
        self.iters = 0
        self.last_norm = float('nan')

    def __call__(self, model, real, fake):
        if self.iters % 4 == 3:
            real = real.detach()
            fake = fake.detach()
            gp, g_norm = zero_gp(model, real, fake)
            # Sync the gradient on the next backward
            if not torch.any(torch.isnan(gp)):
                (4 * self.gamma * gp).backward()
            self.last_norm = g_norm
        self.iters += 1
        return self.last_norm


@tu.experimental
def StyleGAN2Recipe(G: nn.Module,
                    D: nn.Module,
                    dataloader,
                    noise_size: int,
                    gpu_id: int,
                    total_num_gpus: int,
                    *,
                    G_lr: float = 2e-3,
                    D_lr: float = 2e-3,
                    tag: str = 'model',
                    ada: bool = True):
    """
    StyleGAN2 Recipe distributed with DistributedDataParallel

    Args:
        G (nn.Module): a Generator.
        D (nn.Module): a Discriminator.
        dataloader: a dataloader conforming to torchvision's API.
        noise_size (int): the size of the input noise vector.
        gpu_id (int): the GPU index on which to run.
        total_num_gpus (int): how many GPUs are they
        G_lr (float): RAdamW lr for G
        D_lr (float): RAdamW lr for D
        tag (str): tag for Visdom and checkpoints
        ada (bool): whether to enable Adaptive Data Augmentation

    Returns:
        recipe, G EMA model
    """
    G_polyak = copy.copy(G)

    G = nn.parallel.DistributedDataParallel(G.to(gpu_id), [gpu_id], gpu_id)
    D = nn.parallel.DistributedDataParallel(D.to(gpu_id), [gpu_id], gpu_id)
    print(G)
    print(D)

    optG: torch.optim.Optimizer = RAdamW(G.parameters(),
                                         G_lr,
                                         betas=(0., 0.99),
                                         weight_decay=0)
    optD: torch.optim.Optimizer = RAdamW(D.parameters(),
                                         D_lr,
                                         betas=(0., 0.99),
                                         weight_decay=0)

    batch_size = len(next(iter(dataloader))[0])
    diffTF = ADATF(-2 if not ada else -0.9,
                   50000 / (batch_size * total_num_gpus))

    ppl = PPL(4)

    def G_train(batch):
        with G.no_sync():
            pl = ppl(G, torch.randn(batch_size, noise_size, device=gpu_id))
        ##############
        #   G pass   #
        ##############
        imgs = G(torch.randn(batch_size, noise_size, device=gpu_id))
        pred = D(diffTF(imgs) * 2 - 1)
        score = gan_loss.generated(pred)
        score.backward()

        return {'G_loss': score.item(), 'ppl': pl}

    gradient_penalty = GradientPenalty(0.1)

    def D_train(batch):
        ###################
        #    Fake pass    #
        ###################
        with D.no_sync():
            # Sync the gradient on the last backward
            noise = torch.randn(batch_size, noise_size, device=gpu_id)
            with torch.no_grad():
                fake = G(noise)
            fake.requires_grad_(True)
            fake.retain_grad()
            fake_tf = diffTF(fake) * 2 - 1
            fakeness = D(fake_tf).squeeze(1)
            fake_loss = gan_loss.fake(fakeness)
            fake_loss.backward()

            correct = (fakeness < 0).int().eq(1).float().sum()
        fake_grad = fake.grad.detach().norm(dim=1, keepdim=True)
        fake_grad /= fake_grad.max()

        tfmed = diffTF(batch[0]) * 2 - 1

        with D.no_sync():
            grad_norm = gradient_penalty(D, batch[0] * 2 - 1,
                                         fake.detach() * 2 - 1)

        ###################
        #    Real pass    #
        ###################
        real_out = D(tfmed)
        correct += (real_out > 0).detach().int().eq(1).float().sum()
        real_loss = gan_loss.real(real_out)
        real_loss.backward()
        pos_ratio = real_out.gt(0).float().mean().cpu().item()
        diffTF.log_loss(-pos_ratio)
        return {
            'imgs': fake.detach(),
            'i_grad': fake_grad,
            'loss': real_loss.item() + fake_loss.item(),
            'fake_loss': fake_loss.item(),
            'real_loss': real_loss.item(),
            'ADA-p': diffTF.p,
            'D-correct': correct / (2 * real_out.numel()),
            'grad_norm': grad_norm
        }

    tu.freeze(G_polyak)

    def test(batch):
        G_polyak.eval()

        def sample(N, n_iter, alpha=0.01, show_every=10):
            noise = torch.randn(N,
                                noise_size,
                                device=gpu_id,
                                requires_grad=True)
            opt = torch.optim.Adam([noise], lr=alpha)
            fakes = []
            for i in range(n_iter):
                noise += torch.randn_like(noise) / 10
                fake_batch = []
                opt.zero_grad()
                for j in range(0, N, batch_size):
                    with torch.enable_grad():
                        n_batch = noise[j:j + batch_size]
                        fake = G_polyak(n_batch, mixing=False)
                        fake_batch.append(fake)
                        log_prob = n_batch[:, 32:].pow(2).mul_(-0.5)
                        fakeness = -D(fake * 2 - 1).sum() - log_prob.sum()
                        fakeness.backward()
                opt.step()
                fake_batch = torch.cat(fake_batch, dim=0)

                if i % show_every == 0:
                    fakes.append(fake_batch.cpu().detach().clone())

            fakes.append(fake_batch.cpu().detach().clone())

            return torch.cat(fakes, dim=0)

        fake = sample(8, 50, alpha=0.001, show_every=10)

        noise1 = torch.randn(batch_size * 2 // 8, 1, noise_size, device=gpu_id)
        noise2 = torch.randn(batch_size * 2 // 8, 1, noise_size, device=gpu_id)
        t = torch.linspace(0, 1, 8, device=noise1.device).view(8, 1)
        noise = noise1 * t + noise2 * (1 - t)
        noise = noise.view(-1, noise_size)
        interp = torch.cat([
            G_polyak(n, mixing=False) for n in torch.split(noise, batch_size)
        ], dim=0)
        return {
            'polyak_imgs': fake,
            'polyak_interp': interp,
        }

    recipe = GANRecipe(G,
                       D,
                       G_train,
                       D_train,
                       test,
                       dataloader,
                       visdom_env=tag if gpu_id == 0 else None,
                       log_every=10,
                       test_every=1000,
                       checkpoint=tag if gpu_id == 0 else None,
                       g_every=1)
    recipe.callbacks.add_callbacks([
        tcb.Log('batch.0', 'x'),
        tcb.WindowedMetricAvg('fake_loss'),
        tcb.WindowedMetricAvg('real_loss'),
        tcb.WindowedMetricAvg('grad_norm'),
        tcb.WindowedMetricAvg('ADA-p'),
        tcb.WindowedMetricAvg('D-correct'),
        tcb.Log('i_grad', 'img_grad'),
        tch.callbacks.Optimizer(optD),
    ])
    recipe.G_loop.callbacks.add_callbacks([
        tch.callbacks.Optimizer(optG),
        tcb.Polyak(G.module, G_polyak,
                   0.5**((batch_size * total_num_gpus) / 20000)),
        tcb.WindowedMetricAvg('ppl'),
    ])
    recipe.test_loop.callbacks.add_callbacks([
        tcb.Log('polyak_imgs', 'polyak'),
        tcb.Log('polyak_interp', 'interp'),
    ])
    recipe.register('G_polyak', G_polyak)
    recipe.to(gpu_id)
    return recipe, G_polyak


@tu.experimental
def train(rank, world_size):
    from torchelie.models import StyleGAN2Generator, StyleGAN2Discriminator
    from torchvision.datasets import ImageFolder
    import torchvision.transforms as TF
    import argparse

    parser = argparse.ArgumentParser()
    # parser.add_argument('--device', default='cuda')
    parser.add_argument('--batch-size', type=int, default=64)
    parser.add_argument('--noise-size', type=int, default=128)
    parser.add_argument('--img-size', type=int, default=64)
    parser.add_argument('--img-dir', required=True)
    parser.add_argument('--ch-mul', type=float, default=1.)
    parser.add_argument('--max-ch', type=int, default=512)
    parser.add_argument('--no-ada', action='store_true')
    parser.add_argument('--gp-thresh', type=float, default=0.3)
    parser.add_argument('--from-ckpt')
    opts = parser.parse_args()

    tag = opts.img_dir.split('/')[-1] or opts.img_dir.split('/')[-2]
    tag = 'gan_' + tag

    G = StyleGAN2Generator(opts.noise_size,
                           img_size=opts.img_size,
                           ch_mul=opts.ch_mul,
                           max_ch=opts.max_ch)
    D = StyleGAN2Discriminator(input_sz=opts.img_size,
                               ch_mul=opts.ch_mul,
                               max_ch=opts.max_ch)

    tfm = TF.Compose([
        tch.transforms.ResizeNoCrop(opts.img_size),
        tch.transforms.AdaptPad((opts.img_size, opts.img_size),
                                padding_mode='edge'),
        TF.Resize(opts.img_size),
        TF.RandomHorizontalFlip(),
        TF.ToTensor()
    ])
    ds = tch.datasets.NoexceptDataset(ImageFolder(opts.img_dir, transform=tfm))
    dl = torch.utils.data.DataLoader(ds,
                                     num_workers=4,
                                     shuffle=True,
                                     pin_memory=True,
                                     drop_last=True,
                                     batch_size=opts.batch_size)
    recipe, _ = StyleGAN2Recipe(G,
                                D,
                                dl,
                                opts.noise_size,
                                rank,
                                world_size,
                                tag=tag,
                                ada=not opts.no_ada)
    if opts.from_ckpt is not None:
        ckpt = torch.load(opts.from_ckpt, map_location='cpu')
        recipe.load_state_dict(ckpt)

    recipe.run(5000)


if __name__ == '__main__':
    tu.parallel_run(train)



import torch
from torchelie.models import *


def test_patchgan():
    for M in [patch286, patch70, patch34, patch16]:
        m = M()
        m(torch.randn(1, 3, 128, 128))


def test_pnet():
    pnet = PerceptualNet(['conv5_2'])
    pnet(torch.randn(1, 3, 128, 128), detach=True)


def test_factored_predictor():
    fp = PixelPredictor(10)
    fp(torch.randn(5, 10), torch.randn(5, 3))
    fp.sample(torch.randn(3, 10), 1)


def test_pixelcnn():
    pc = PixelCNN(10, (8, 8), 1)
    pc(torch.randn(1, 1, 8, 8))
    pc.sample(1, 1)

    pc = PixelCNN(10, (8, 8), 3)
    pc(torch.randn(1, 3, 8, 8))
    pc.sample(1, 1)


def test_resnet():
    m = ResidualDiscriminator([2, 'D', 3])
    m(torch.randn(1, 3, 8, 8))

    m = ResidualDiscriminator([2, 'D', 3])
    m.to_projection_discr(3)
    m(torch.randn(1, 3, 8, 8), torch.LongTensor([1]))

    def run(M):
        m = M(4)
        out = m(torch.randn(2, 3, 32, 32))
        out.mean().backward()

    run(resnet18)
    run(preact_resnet18)
    run(resnet50)
    run(preact_resnet50)
    run(resnext50_32x4d)
    run(preact_resnext50_32x4d)


def test_unet():
    m = UNet([3, 6, 12], 1)
    m(torch.randn(1, 3, 128, 128))


def test_vgg():
    m = vgg11(2)
    m(torch.randn(1, 3, 32, 32))


def test_attention():
    m = attention56(2)
    m(torch.randn(2, 3, 32, 32))


def test_hourglass():
    m = Hourglass()
    out = m(torch.randn(2, 32, 128, 128))
    assert out.shape == (2, 3, 128, 128)


def test_autogan():
    m = AutoGAN([3, 4, 5], in_noise=4)
    m(torch.randn(16, 4))



import torch
import torchelie as tch
import torch.nn as nn
from torch.utils.data import DataLoader
from torchvision.transforms import ToPILImage

from torchelie.recipes import CrossEntropyClassification
from torchelie.recipes import DeepDream
from torchelie.recipes import FeatureVis
from torchelie.recipes import NeuralStyle
from torchelie.recipes import TrainAndCall
import torchelie.callbacks as tcb


class FakeData:

    def __len__(self):
        return 10

    def __getitem__(self, i):
        cls = 0 if i < 5 else 1
        return torch.randn(10) + cls * 3, cls


class FakeImg:

    def __len__(self):
        return 10

    def __getitem__(self, i):
        cls = 0 if i < 5 else 1
        return torch.randn(1, 4, 4) + cls * 3, cls


def test_classification():
    trainloader = DataLoader(FakeImg(), 4, shuffle=True)
    testloader = DataLoader(FakeImg(), 4, shuffle=True)

    model = nn.Sequential(tch.nn.Reshape(-1), nn.Linear(16, 2))

    clf_recipe = CrossEntropyClassification(model, trainloader, testloader,
                                            ['foo', 'bar'])
    clf_recipe.run(1)


def test_deepdream():
    model = nn.Sequential(nn.Conv2d(3, 6, 3))
    dd = DeepDream(model, '0')
    dd.fit(ToPILImage()(torch.randn(3, 128, 128)), 1)


def test_featurevis():
    model = nn.Sequential(nn.Conv2d(3, 6, 3))
    dd = FeatureVis(model, '0', 229, lr=1)
    dd.fit(1, 0)


def test_neuralstyle():
    stylizer = NeuralStyle()

    content = ToPILImage()(torch.randn(3, 64, 64))
    style_img = ToPILImage()(torch.randn(3, 64, 64))

    stylizer.fit(1,
                 content,
                 style_img,
                 1,
                 second_scale_ratio=1,
                 content_layers=['conv1_1'])


def test_trainandcall():
    model = nn.Linear(10, 2)

    def train_step(batch):
        x, y = batch
        out = model(x)
        loss = torch.nn.functional.cross_entropy(out, y)
        loss.backward()
        return {'loss': loss}

    def after_train():
        print('Yup.')
        return {}

    trainloader = DataLoader(FakeData(), 4, shuffle=True)
    trainer = TrainAndCall(model, train_step, after_train, trainloader)
    trainer.callbacks.add_callbacks(
        [tcb.Optimizer(torch.optim.Adam(model.parameters(), lr=1e-3))])

    trainer.run(1)


def test_callbacks():
    from torchelie.recipes import Recipe

    def train(b):
        x, y = b
        return {'pred': torch.randn(y.shape[0])}

    m = nn.Linear(2, 2)

    r = Recipe(train, DataLoader(FakeImg(), 4))
    r.callbacks.add_callbacks([
        tcb.Counter(),
        tcb.AccAvg(),
        tcb.Checkpoint('/tmp/m', m),
        tcb.ClassificationInspector(1, ['1', '2']),
        tcb.ConfusionMatrix(['one', 'two']),
        tcb.ImageGradientVis(),
        tcb.MetricsTable(),
    ])
    r.run(1)



from torchelie.optim import *


def test_deepdream():
    a = torch.randn(5, requires_grad=True)
    opt = DeepDreamOptim([a])
    a.mean().backward()
    opt.step()


def test_addsign():
    a = torch.randn(5, requires_grad=True)
    opt = AddSign([a])
    a.mean().backward()
    opt.step()


def test_radamw():
    a = torch.randn(5, requires_grad=True)
    opt = AddSign([a])
    a.mean().backward()
    opt.step()


def test_lookahead():
    a = torch.randn(5, requires_grad=True)
    opt = Lookahead(AddSign([a]))
    a.mean().backward()
    opt.step()



import torch.nn as nn
from typing import Any, Optional, Callable, cast


class CondSeq(nn.Sequential):
    """
    An extension to torch's Sequential that allows conditioning either as a
    second forward argument or `condition()`
    """
    def condition(self, z: Any) -> None:
        """
        Conditions all the layers on z

        Args:
            z: conditioning
        """
        for m in self:
            if hasattr(m, 'condition') and m is not self:
                cast(Callable, m.condition)(z)

    def forward(self, x: Any, z: Optional[Any] = None) -> Any:
        """
        Forward pass

        Args:
            x: input
            z (optional): conditioning. condition() must be called first if
                left None
        """
        for m in self:
            if hasattr(m, 'condition') and z is not None:
                x = m(x, z)
            else:
                x = m(x)
        return x



import torch
import torch.nn as nn
import torch.nn.functional as F

from typing import Union, Tuple

from .functional import quantize


class VQ(nn.Module):
    """
    Quantization layer from *Neural Discrete Representation Learning*

    Args:
        latent_dim (int): number of features along which to quantize
        num_tokens (int): number of tokens in the codebook
        dim (int): dimension along which to quantize
        mode ('angular' or 'nearest'): whether the distance between the input
            vectors and the codebook vectors is computed with a L2 distance or
            an angular distance
        return_indices (bool): whether to return the indices of the quantized
            code points
    """
    embedding: nn.Embedding
    dim: int
    commitment: float
    initialized: torch.Tensor
    mode: str
    return_indices: bool
    init_mode: str

    def __init__(self,
                 latent_dim: int,
                 num_tokens: int,
                 dim: int = 1,
                 commitment: float = 0.25,
                 mode: str = 'nearest',
                 init_mode: str = 'normal',
                 return_indices: bool = True):
        super(VQ, self).__init__()
        self.embedding = nn.Embedding(num_tokens, latent_dim)
        nn.init.normal_(self.embedding.weight, 0, 1)
        self.dim = dim
        self.commitment = commitment
        self.register_buffer('initialized', torch.Tensor([0]))
        assert mode in ['nearest', 'angular']
        self.mode = mode
        self.return_indices = return_indices
        assert init_mode in ['normal', 'first']
        self.init_mode = init_mode

    def forward(
        self, x: torch.Tensor
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """
        Forward pass

        Args:
            x (tensor): input tensor

        Returns:
            quantized tensor, or (quantized tensor, indices) if
            `self.return_indices`
        """
        dim = self.dim
        nb_codes = self.embedding.weight.shape[0]

        codebook = self.embedding.weight
        if self.mode == 'angular':
            codebook = F.normalize(codebook)
            x = F.normalize(x, dim=dim)

        if (self.init_mode == 'first' and self.initialized.item() == 0
                and self.training):
            n_proto = self.embedding.weight.shape[0]

            ch_first = x.transpose(dim, -1).contiguous().view(-1, x.shape[dim])
            n_samples = ch_first.shape[0]
            idx = torch.randint(0, n_samples, (n_proto, ))[:nb_codes]
            self.embedding.weight.data.copy_(ch_first[idx])
            self.initialized[:] = 1

        needs_transpose = dim != -1 or dim != x.dim() - 1
        if needs_transpose:
            x = x.transpose(-1, dim).contiguous()

        codes, indices = quantize(x, codebook, self.commitment, self.dim)

        if needs_transpose:
            codes = codes.transpose(-1, dim)
            indices = indices.transpose(-1, dim)

        if self.return_indices:
            return codes, indices
        else:
            return codes


class MultiVQ(nn.Module):
    """
    Multi codebooks quantization layer from *Neural Discrete Representation
    Learning*

    Args:
        latent_dim (int): number of features along which to quantize
        num_tokens (int): number of tokens in the codebook
        num_codebooks (int): number of parallel codebooks
        dim (int): dimension along which to quantize
        mode ('angular' or 'nearest'): whether the distance between the input
            vectors and the codebook vectors is computed with a L2 distance or
            an angular distance
        return_indices (bool): whether to return the indices of the quantized
            code points
    """
    def __init__(self,
                 latent_dim: int,
                 num_tokens: int,
                 num_codebooks: int,
                 dim: int = 1,
                 commitment: float = 0.25,
                 mode: str = 'nearest',
                 init_mode: str = 'normal',
                 return_indices: bool = True):
        assert latent_dim % num_codebooks == 0, (
            "num_codebooks must divide evenly latent_dim")
        super(MultiVQ, self).__init__()
        self.dim = dim
        self.num_codebooks = num_codebooks
        self.vqs = nn.ModuleList([
            VQ(latent_dim // num_codebooks,
               num_tokens,
               dim=dim,
               commitment=commitment,
               mode=mode,
               init_mode=init_mode,
               return_indices=return_indices) for _ in range(num_codebooks)
        ])

    def forward(
        self, x: torch.Tensor
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        x_chunks = torch.chunk(x, self.num_codebooks, dim=self.dim)
        return torch.cat([vq(chunk) for chunk, vq in zip(x_chunks, self.vqs)],
                         dim=self.dim)



import torch
import torch.nn as nn
import torch.nn.functional as F

from .conv import Conv3x3
from torchelie.utils import kaiming, xavier, experimental

__all__ = []


class BatchNorm2dBase_(nn.Module):
    @experimental
    def __init__(self, channels, momentum=0.8):
        super(BatchNorm2dBase_, self).__init__()
        self.register_buffer('running_mean', torch.zeros(1, channels, 1, 1))
        self.register_buffer('running_var', torch.ones(1, channels, 1, 1))
        self.register_buffer('step', torch.ones(1))
        self.momentum = momentum

    def update_moments(self, x):
        if self.training:
            m = x.mean(dim=(0, 2, 3), keepdim=True)
            v = torch.sqrt(x.var(dim=(0, 2, 3), unbiased=False, keepdim=True) + 1e-8)

            self.running_mean.copy_((self.momentum * self.running_mean
                                     + (1 - self.momentum) * m))

            self.running_var.copy_((self.momentum * self.running_var
                                    + (1 - self.momentum) * v))

            self.step += 1
        else:
            m = self.running_mean
            v = self.running_var
        return m, v


__all__.append('BatchNorm2dBase_')


class MovingAverageBN2dBase_(nn.Module):
    @experimental
    def __init__(self, channels, momentum=0.8):
        super(MovingAverageBN2dBase_, self).__init__()
        self.register_buffer('running_mean', torch.zeros(1, channels, 1, 1))
        self.register_buffer('running_var', torch.zeros(1, channels, 1, 1))
        self.register_buffer('step', torch.ones(1))
        self.momentum = momentum

    def update_moments(self, x):
        if self.training:
            m = x.mean(dim=(0, 2, 3), keepdim=True)
            v = x.var(dim=(0, 2, 3), keepdim=True)

            m = self.momentum * self.running_mean + (1 - self.momentum) * m
            self.running_mean.copy_(m.detach())
            m = m / (1 - self.momentum**self.step)

            v = self.momentum * self.running_var + (1 - self.momentum) * v
            self.running_var.copy_(v.detach())
            v = v / (1 - self.momentum**self.step)

            self.step += 1
        else:
            m = self.running_mean
            v = self.running_var
        return m, torch.sqrt(v)


__all__.append('MovingAverageBN2dBase_')


def make_no_affine(base, name):
    class NoAffineBN(base):
        @experimental
        def __init__(self, channels, momentum=0.8):
            super(NoAffineBN, self).__init__(channels, momentum)

        def forward(self, x):
            m, s = self.update_moments(x)
            return (x - m) / (s + 1e-8)

    NoAffineBN.__name__ = name
    return NoAffineBN


NoAffineBN2d = make_no_affine(BatchNorm2dBase_, 'NoAffineBN2d')
NoAffineMABN2d = make_no_affine(MovingAverageBN2dBase_, 'NoAffineMABN2d')
__all__.append('NoAffineBN2d')
__all__.append('NoAffineMABN2d')


def make_bn(base, name):
    class BatchNorm2d(base):
        @experimental
        def __init__(self, channels, momentum=0.8):
            super(BatchNorm2d, self).__init__(channels, momentum)
            self.weight = nn.Parameter(torch.zeros(channels, 1, 1))
            self.bias = nn.Parameter(torch.zeros(channels, 1, 1))

        def forward(self, x):
            m, s = self.update_moments(x)
            weight = (1 + self.weight) / (s + 1e-8)
            bias = -m * weight + self.bias
            return weight * x + bias

    BatchNorm2d.__name__ = name
    return BatchNorm2d


BatchNorm2d = make_bn(BatchNorm2dBase_, 'BatchNorm2d')
MovingAverageBN2d = make_bn(MovingAverageBN2dBase_, 'MovingAverageBN2d')
__all__.append('BatchNorm2d')
__all__.append('MovingAverageBN2d')


def make_cbn(base, name):
    class ConditionalBN2d(base):
        @experimental
        def __init__(self, channels, cond_channels, momentum=0.8):
            super(ConditionalBN2d, self).__init__(channels, momentum)
            self.make_weight = nn.Linear(cond_channels, channels)
            nn.init.normal_(self.make_weight.weight, 0.002)
            nn.init.zeros_(self.make_weight.bias)
            self.make_bias = nn.Linear(cond_channels, channels)
            nn.init.normal_(self.make_bias.weight, 0.002)
            nn.init.zeros_(self.make_bias.bias)
            self.bn = nn.BatchNorm2d(channels, affine=False)

            self.weight = torch.tensor([0])
            self.bias = torch.tensor([0])

        def forward(self, x, z=None):
            if z is not None:
                self.condition(z)
            return self.bn(x) * self.weight + self.bias

            m, v = self.update_moments(x)
            weight = self.weight / (v + 1e-6)
            bias = -m * weight + self.bias
            out = x * weight + bias
            print(out.mean().item(), out.std().item())
            return out

        def condition(self, z):
            self.weight = (self.make_weight(z)[:, :, None, None])+1
            self.bias = (self.make_bias(z)[:, :, None, None])

    ConditionalBN2d.__name__ = name
    return ConditionalBN2d


ConditionalBN2d = make_cbn(BatchNorm2dBase_, 'ConditionalBN2d')
ConditionalMABN2d = make_cbn(MovingAverageBN2dBase_, 'ConditionalMABN2d')
__all__.append('ConditionalBN2d')
__all__.append('ConditionalMABN2d')


def make_spade(base, name):
    class Spade2d(base):
        @experimental
        def __init__(self,
                     channels,
                     cond_channels,
                     hidden,
                     momentum=0.8):
            super(Spade2d, self).__init__(channels, momentum)
            self.initial = kaiming(Conv3x3(cond_channels, hidden, stride=2))
            self.make_weight = xavier(Conv3x3(hidden, channels))
            self.make_bias = xavier(Conv3x3(hidden, channels))
            self.register_buffer('weight', torch.ones(channels))

        def forward(self, x, z=None):
            if z is not None:
                self.z = z
            self.make_weight_bias(self.z, x.shape[2:])

            m, v = self.update_moments(x)
            weight = (1 + self.weight) / (v + 1e-8)
            bias = -m * weight + self.bias
            return weight * x + bias

        def condition(self, z):
            self.z = z

        def make_weight_bias(self, z, size):
            z = F.interpolate(z,
                              size=(size[0] * 2, size[1] * 2),
                              mode='nearest')
            z = F.relu(self.initial(z), inplace=True)
            self.weight = self.make_weight(z)
            self.bias = self.make_bias(z)

    Spade2d.__name__ = name
    return Spade2d


Spade2d = make_spade(BatchNorm2dBase_, 'Spade2d')
SpadeMA2d = make_spade(MovingAverageBN2dBase_, 'SpadeMA2d')
__all__.append('Spade2d')
__all__.append('SpadeMA2d')


class AttenNorm2d(nn.BatchNorm2d):
    """
    From https://arxiv.org/abs/1908.01259
    """
    def __init__(self,
                 num_features,
                 num_weights,
                 eps=1e-8,
                 momentum=0.8,
                 track_running_stats=True):
        super(AttenNorm2d, self).__init__(num_features,
                                          eps=eps,
                                          momentum=momentum,
                                          affine=False,
                                          track_running_stats=track_running_stats)
        self.gamma = nn.Parameter(torch.ones(num_weights, num_features))
        self.beta = nn.Parameter(torch.zeros(num_weights, num_features))

        self.avgpool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Linear(num_features, num_weights)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        output = super(AttenNorm2d, self).forward(x)
        size = output.size()
        b, c, _, _ = x.size()

        y = self.avgpool(x).view(b, c)
        y = self.fc(y)
        y = self.sigmoid(y)

        gamma = torch.mm(y, self.gamma)
        beta = torch.mm(y, self.beta)

        gamma = gamma.unsqueeze(-1).unsqueeze(-1).expand(size)
        beta = beta.unsqueeze(-1).unsqueeze(-1).expand(size)
        return gamma * output + beta


__all__.append('AttenNorm2d')



import torch
from torch.autograd import Function


class VectorQuantization(Function):

    @staticmethod
    def compute_indices(inputs_orig, codebook):
        bi = []
        SZ = 10000
        for i in range(0, inputs_orig.size(0), SZ):
            inputs = inputs_orig[i:i + SZ]
            # NxK
            distances_matrix = torch.cdist(inputs, codebook)
            # Nx1
            indic = torch.min(distances_matrix, dim=-1)[1].unsqueeze(1)
            bi.append(indic)
        return torch.cat(bi, dim=0)

    @staticmethod
    def flatten(x):
        code_dim = x.size(-1)
        return x.view(-1, code_dim)

    @staticmethod
    def restore_shapes(codes, indices, target_shape):
        idx_shape = list(target_shape)
        idx_shape[-1] = 1
        return codes.view(*target_shape), indices.view(*idx_shape)

    @staticmethod
    def forward(ctx, inputs, codebook, commitment=0.25, dim=1):
        inputs_flat = VectorQuantization.flatten(inputs)
        indices = VectorQuantization.compute_indices(inputs_flat, codebook)
        codes = codebook[indices.view(-1), :]
        codes, indices = VectorQuantization.restore_shapes(
            codes, indices, inputs.shape)

        ctx.save_for_backward(codes, inputs, torch.tensor([float(commitment)]),
                              codebook, indices)
        ctx.mark_non_differentiable(indices)
        return codes, indices

    @staticmethod
    def backward(ctx, straight_through, unused_indices):
        codes, inputs, beta, codebook, indices = ctx.saved_tensors

        # TODO: figure out proper vq loss reduction
        # vq_loss = F.mse_loss(inputs, codes).detach()

        # gradient of vq_loss
        diff = 2 * (inputs - codes) / inputs.numel()

        commitment = beta.item() * diff

        code_disp = VectorQuantization.flatten(-diff)
        indices = VectorQuantization.flatten(indices)
        code_disp = (torch.zeros_like(codebook).index_add_(
            0, indices.view(-1), code_disp))
        return straight_through + commitment, code_disp, None, None


quantize = VectorQuantization.apply



from typing import List, Tuple, cast

import torchelie.nn as tnn
import torchelie.utils as tu
import torch.nn as nn


@tu.experimental
class UNet(tnn.CondSeq):
    """
    U-Net from `U-Net: Convolutional Networks for Biomedical Image Segmentation
    <https://arxiv.org/abs/1505.04597>`_. This net has architectural changes
    operations for further customization.

    Args:
        arch (List[int]): a list of channels from the outermost to innermost
            layers
        num_classes (int): number of output channels
    """

    def __init__(self, arch: List[int], num_classes: int) -> None:
        super().__init__()
        self.arch = arch
        self.in_channels = 3
        self.out_channels = arch[-1]

        feats = tnn.CondSeq()
        feats.input = tnn.ConvBlock(3, arch[0], 3)

        encdec: nn.Module = tnn.ConvBlock(arch[-1], arch[-1] * 2, 3)
        for outer, inner in zip(arch[-2::-1], arch[:0:-1]):
            encdec = tnn.UBlock(outer, inner, encdec)
        feats.encoder_decoder = encdec
        self.features = feats
        assert isinstance(encdec.out_channels, int)
        self.classifier = tnn.ConvBlock(encdec.out_channels, num_classes,
                                        3).remove_batchnorm().no_relu()

    def leaky(self, leak: float = 0.2) -> 'UNet':
        for m in self.modules():
            if isinstance(m, tnn.ConvBlock):
                m.leaky(leak)
        return self

    def set_padding_mode(self, mode: str) -> 'UNet':
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                m.padding_mode = mode
        return self

    def set_input_specs(self, in_channels: int) -> 'UNet':
        assert isinstance(self.features.input, tnn.ConvBlock)
        c = self.features.input.conv
        self.features.input.conv = tu.kaiming(
            nn.Conv2d(in_channels,
                      c.out_channels,
                      cast(Tuple[int, int], c.kernel_size),
                      bias=c.bias is not None,
                      padding=cast(Tuple[int, int], c.padding)))
        return self

    def set_encoder_num_layers(self, num: int) -> 'UNet':
        for m in self.modules():
            if isinstance(m, tnn.UBlock):
                m.set_encoder_num_layers(num)
        return self

    def set_decoder_num_layers(self, num: int) -> 'UNet':
        for m in self.modules():
            if isinstance(m, tnn.UBlock):
                m.set_decoder_num_layers(num)
        return self

    def to_bilinear_sampling(self) -> 'UNet':
        for m in self.modules():
            if isinstance(m, tnn.UBlock):
                m.to_bilinear_sampling()
        return self

    def remove_first_batchnorm(self) -> 'UNet':
        assert isinstance(self.features.input, tnn.ConvBlock)
        self.features.input.remove_batchnorm()
        return self

    def remove_batchnorm(self) -> 'UNet':
        for m in self.modules():
            if isinstance(m, tnn.ConvBlock):
                m.remove_batchnorm()
        return self



import torch
from torchelie.data_learning import *


def test_pixel_image():
    pi = PixelImage((1, 3, 128, 128), 0.01)
    pi()

    start = torch.randn(3, 128, 128)
    pi = PixelImage((1, 3, 128, 128), init_img=start)

    assert start.allclose(pi() + 0.5, atol=1e-7)


def test_spectral_image():
    pi = SpectralImage((1, 3, 128, 128), 0.01)
    pi()

    start = torch.randn(1, 3, 128, 128)
    pi = SpectralImage((1, 3, 128, 128), init_img=start)


def test_correlate_colors():
    corr = CorrelateColors()
    start = torch.randn(1, 3, 64, 64)
    assert start.allclose(corr.invert(corr(start)), atol=1e-5)


def test_parameterized_img():
    start = torch.clamp(torch.randn(1, 3, 128, 128) + 0.5, min=0, max=1)

    ParameterizedImg(1, 3, 128, 128, space='spectral', colors='uncorr')()
    ParameterizedImg(1, 3,
                     128,
                     128,
                     space='spectral',
                     colors='uncorr',
                     init_img=start)()

    ParameterizedImg(1, 3, 128, 128, space='spectral', colors='uncorr')()

    start = torch.clamp(torch.randn(1, 3, 128, 129) + 0.5, min=0, max=1)
    ParameterizedImg(1, 3,
                     128,
                     129,
                     space='spectral',
                     colors='uncorr',
                     init_img=start)()
    start = torch.clamp(torch.randn(1, 3, 128, 128) + 0.5, min=0, max=1)
    ParameterizedImg(1, 3, 128, 128, space='pixel', colors='uncorr')()
    ParameterizedImg(1, 3,
                     128,
                     128,
                     space='pixel',
                     colors='uncorr',
                     init_img=start)()

    ParameterizedImg(1, 3, 128, 128, space='spectral', colors='corr')()
    ParameterizedImg(1, 3,
                     128,
                     128,
                     space='spectral',
                     colors='corr',
                     init_img=start)()

    ParameterizedImg(1, 3, 128, 128, space='pixel', colors='corr')()
    ParameterizedImg(1, 3, 128, 128, space='pixel', colors='corr',
                     init_img=start)()



import torch
from typing import List, cast

import torchelie.nn as tnn
import torchelie.utils as tu
import torch.nn as nn
from .unet import UNet


class Pix2PixGenerator(UNet):
    """
    UNet generator from Pix2Pix. Dropout layers have been substitued with Noise
    injections from StyleGAN2.

    Args:
        arch (List[int]): the number of channel for each depth level of the
            UNet.
    """

    def __init__(self, arch: List[int]) -> None:
        super().__init__(arch, 3)
        self.remove_first_batchnorm()

        self.features.input = tnn.ConvBlock(3, int(arch[0]), 7)
        self.features.input.remove_batchnorm()

        encdec = cast(nn.Module, self.features.encoder_decoder)
        for m in encdec.modules():
            if isinstance(m, tnn.UBlock):
                m.to_bilinear_sampling()
                m.set_encoder_num_layers(1)
                m.set_decoder_num_layers(1)

        tnn.utils.make_leaky(self)
        self.classifier.relu = nn.Sigmoid()

        self._add_noise()

    def _add_noise(self):
        layers = [self.features.encoder_decoder]
        while hasattr(layers[-1].inner, 'inner'):
            layers.append(layers[-1].inner)
        tnn.utils.insert_after(layers[-1].inner, 'norm', tnn.Noise(1, True),
                               'noise')

        for m in layers:
            tnn.utils.insert_after(
                m.out_conv.conv_0, 'norm',
                tnn.Noise(m.out_conv.conv_0.out_channels, True), 'noise')

    def to_equal_lr(self) -> 'Pix2PixGenerator':
        return tnn.utils.net_to_equal_lr(self)

    def set_padding_mode(self, mode: str) -> 'Pix2PixGenerator':
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                m.padding_mode = mode
        return self

    @torch.no_grad()
    def to_instance_norm(self, affine: bool = True) -> 'Pix2PixGenerator':
        """
        Pix2Pix sometimes uses batch size 1, similar to instance norm.
        """

        def to_instancenorm(m):
            if isinstance(m, nn.BatchNorm2d):
                return nn.InstanceNorm2d(m.num_features, affine=affine)
            return m

        tnn.utils.edit_model(self, to_instancenorm)

        return self


def pix2pix_256() -> Pix2PixGenerator:
    """
    The architecture used in `Pix2Pix <https://arxiv.org/abs/1611.07004>`_,
    able to train on 256x256 or 512x512 images.
    """
    return Pix2PixGenerator([64, 128, 256, 512, 512, 512, 512, 512])


def pix2pix_dev() -> Pix2PixGenerator:
    """
    A version of pix2pix_256 with less filter to use less memory and compute.
    """
    return Pix2PixGenerator([32, 64, 128, 128, 256, 256, 512, 512])



from .classifier import ClassificationHead, ProjectionDiscr
from .vgg import *
from .resnet import *
from .patchgan import *
from .pixcnn import *
from .perceptualnet import *
from .unet import *
from .pix2pix import *
from .pix2pixhd import *
from .autogan import AutoGAN, autogan_32, autogan_64, autogan_128
from .snres_discr import *
from .hourglass import Hourglass
from .attention import Attention56Bone, attention56
from .stylegan2 import StyleGAN2Generator, StyleGAN2Discriminator
from .efficient import EfficientNet



from collections import OrderedDict

import torch.nn as nn
import torchvision.models as M
from torchelie.nn import WithSavedActivations
from typing import List


class PerceptualNet(WithSavedActivations):
    """
    Make a VGG16 with appropriately named layers that records intermediate
    activations.

    Args:
        layers (list of str): the names of the layers for which to save the
            activations.
        use_avg_pool (bool): Whether to replace max pooling with averange
            pooling (default: True)
        remove_unused_layers (bool): whether to remove layers past the last one
            used (default: True)
    """
    def __init__(self, layers: List[str],
                 use_avg_pool: bool = True,
                 remove_unused_layers: bool = True) -> None:
        # yapf: disable
        layer_names = [
            'conv1_1', 'relu1_1', 'conv1_2', 'relu1_2', 'maxpool1',
            'conv2_1', 'relu2_1', 'conv2_2', 'relu2_2', 'maxpool2',
            'conv3_1', 'relu3_1', 'conv3_2', 'relu3_2', 'conv3_3', 'relu3_3',
                    'conv3_4', 'relu3_4', 'maxpool3',  # noqa: E131
            'conv4_1', 'relu4_1', 'conv4_2', 'relu4_2', 'conv4_3', 'relu4_3',
                    'conv4_4', 'relu4_4', 'maxpool4',  # noqa: E131
            'conv5_1', 'relu5_1', 'conv5_2', 'relu5_2', 'conv5_3', 'relu5_3',
                    'conv5_4', 'relu5_4',  # 'maxpool5'
        ]
        # yapf: enable

        m = M.vgg19(pretrained=True).eval().features
        m = nn.Sequential(
            OrderedDict([(l_name, l) for l_name, l in zip(layer_names, m)]))
        for nm, mod in m.named_modules():
            if 'relu' in nm:
                setattr(m, nm, nn.ReLU(False))
            elif 'pool' in nm and use_avg_pool:
                setattr(m, nm, nn.AvgPool2d(2, 2))

        if remove_unused_layers:
            m = m[:max([layer_names.index(layer) for layer in layers]) + 1]

        super().__init__(m, names=layers)



import os
import math
import torch.distributed as dist
import torch
import torch.nn as nn
import torch.nn.functional as F
import warnings
from typing import Optional, Iterable, TypeVar, Any, overload
from typing import List
from functools import wraps
from inspect import isfunction
from textwrap import dedent


T = TypeVar('T')
T_Module = TypeVar('T_Module', bound=nn.Module)
Numeric = TypeVar('Numeric', torch.Tensor, float)


def fast_zero_grad(net: nn.Module) -> None:
    """
    Set :code:`.grad` to None for all parameters instead of zeroing out. It is
    faster.
    """
    for p in net.parameters():
        p.grad = None


def freeze(net: T_Module) -> T_Module:
    """
    Freeze all parameters of `net`
    """
    for p in net.parameters():
        p.requires_grad_(False)
    return net


def unfreeze(net: T_Module) -> T_Module:
    """
    Unfreeze all parameters of `net`
    """
    for p in net.parameters():
        p.requires_grad_(True)
    return net


def entropy(out: torch.Tensor,
            dim: int = 1,
            reduce: str = 'mean') -> torch.Tensor:
    """
    Compute the entropy of the categorical distribution specified by the logits
    `out` along dimension `dim`.

    Args:
        out (tensor): logits categorial distribution
        dim (int): the dimension along which the distributions are specified
        reduce (str): `"mean"`, `"none"` or `"sum"`
    """
    log_prob = F.log_softmax(out, dim=dim)
    h = -torch.sum(log_prob.exp() * log_prob, dim=dim)
    if reduce == 'none':
        return h
    if reduce == 'mean':
        return h.mean()
    if reduce == 'sum':
        return h.sum()
    assert False, reduce + ' is not a valid reduction method'


def kaiming_gain(m: T_Module,
                 a: float = 0,
                 nonlinearity='leaky_relu',
                 mode='fan_in') -> float:
    """
    Return the std needed to initialize a weight matrix with given parameters.
    """
    fan = nn.init._calculate_correct_fan(m.weight, mode)
    gain = nn.init.calculate_gain(nonlinearity, param=a)
    return gain / math.sqrt(fan)


def kaiming(m: T_Module,
            a: float = 0,
            nonlinearity: str = 'leaky_relu',
            mode: str = 'fan_in',
            dynamic: bool = False) -> T_Module:
    """
    Initialize a module with kaiming normal init

    Args:
        m (nn.Module): the module to init
        a (float): the slope of the nonlinearity
        nonlinearity (str): type of the nonlinearity
        dynamic (bool): wether to scale the weights on the forward pass for
            equalized LR such as ProGAN (default: False)

    Returns:
        the initialized module
    """
    assert isinstance(m.weight, torch.Tensor)
    if nonlinearity in ['relu', 'leaky_relu']:
        if a == 0:
            nonlinearity = 'relu'
        else:
            nonlinearity = 'leaky_relu'

    if not dynamic:
        nn.init.kaiming_normal_(m.weight,
                                a=a,
                                nonlinearity=nonlinearity,
                                mode=mode)
    else:
        from .nn.utils import weight_scale
        nn.init.normal_(m.weight, 0, 1)
        weight_scale(m,
                     scale=kaiming_gain(m,
                                        a=a,
                                        nonlinearity=nonlinearity,
                                        mode=mode))

    if hasattr(m, 'biais') and m.bias is not None:
        assert isinstance(m.bias, torch.Tensor)
        nn.init.constant_(m.bias, 0)
    return m


def xavier(m: T_Module,
           a: float = 0,
           nonlinearity: str = 'relu',
           mode: str = 'fan_in',
           dynamic: bool = False) -> T_Module:
    """
    Initialize a module with xavier normal init

    Args:
        m (nn.Module): the module to init
        dynamic (bool): wether to scale the weights on the forward pass for
            equalized LR such as ProGAN (default: False)

    Returns:
        the initialized module
    """
    assert isinstance(m.weight, torch.Tensor)
    if nonlinearity in ['relu', 'leaky_relu']:
        if a == 0:
            nonlinearity = 'relu'
        else:
            nonlinearity = 'leaky_relu'

    if not dynamic:
        nn.init.xavier_normal_(m.weight)
    else:
        from .nn.utils import weight_scale
        nn.init.normal_(m.weight, 0, 1)
        fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(m.weight)
        gain = nn.init.calculate_gain(nonlinearity, param=a)
        weight_scale(m, scale=gain * math.sqrt(2. / (fan_in + fan_out)))

    if hasattr(m, 'biais') and m.bias is not None:
        assert isinstance(m.bias, torch.Tensor)
        nn.init.constant_(m.bias, 0)
    return m


def normal_init(m: nn.Module, std: float = 0.02) -> nn.Module:
    """
    Initialize a module with gaussian weights of standard deviation std

    Args:
        m (nn.Module): the module to init

    Returns:
        the initialized module
    """
    assert isinstance(m.weight, torch.Tensor)
    nn.init.normal_(m.weight, 0, std)
    if hasattr(m, 'biais') and m.bias is not None:
        assert isinstance(m.bias, torch.Tensor)
        nn.init.constant_(m.bias, 0)
    return m


def constant_init(m: nn.Module, val: float) -> nn.Module:
    """
    Initialize a module with gaussian weights of standard deviation std

    Args:
        m (nn.Module): the module to init

    Returns:
        the initialized module
    """
    assert isinstance(m.weight, torch.Tensor)
    nn.init.constant_(m.weight, val)
    if hasattr(m, 'biais') and m.bias is not None:
        assert isinstance(m.bias, torch.Tensor)
        nn.init.constant_(m.bias, 0)
    return m


def nb_parameters(net: nn.Module) -> int:
    """
    Counts the number of parameters of `net`

    Args:
        net (nn.Module): the net

    Returns:
        the number of params
    """
    return sum(p.numel() for p in net.parameters())


def layer_by_name(net: nn.Module, name: str) -> Optional[nn.Module]:
    """
    Get a submodule at any depth of a net by its name

    Args:
        net (nn.Module): the base module containing other modules
        name (str): a name of a submodule of `net`, like `"layer3.0.conv1"`.

    Returns:
        The found layer or `None`
    """
    for layer in net.named_modules():
        if layer[0] == name:
            return layer[1]
    return None


def forever(iterable: Iterable[T]) -> Iterable[T]:
    """
    Cycle through `iterable` forever

    Args:
        iterable (iterable): the iterable
    """
    it = iter(iterable)
    while True:
        try:
            yield next(it)
        except Exception as e:
            print(e)
            it = iter(iterable)


def gram(m: torch.Tensor) -> torch.Tensor:
    """
    Return the Gram matrix of `m`

    Args:
        m (tensor): a matrix of dim 2

    Returns:
        The Gram matrix
    """
    g = torch.einsum('ik,jk->ij', m, m) / m.shape[1]
    return g


def bgram(m: torch.Tensor) -> torch.Tensor:
    """
    Return the batched Gram matrix of `m`

    Args:
        m (tensor): a matrix of dim 3, first one is the batch

    Returns:
        The batch of Gram matrix
    """
    m = m.view(m.shape[0], m.shape[1], -1)
    g = torch.einsum('bik,bjk->bij', m, m) / (m.shape[1] * m.shape[2])
    return g


def dict_by_key(d: Any, k: str) -> Any:
    """
    Recursively index a `dict` by a hierarchical key

    ```
    >>> dict_by_key({'a': [{'b': 42}]}, 'a.0.b')
    42
    ```

    Args:
        d (dict, list, and any level of nesting): the data to index
        k (str): the key

    Returns:
        The value in `d` indexed by `k`
    """
    ks = k.split('.')
    while len(ks) != 0:
        if isinstance(d, dict):
            d = d[ks[0]]
        else:
            d = d[int(ks[0])]
        ks = ks[1:]
    return d


def send_to_device(x: Any, device, non_blocking: bool = False) -> Any:
    """
    Send all tensors contained in `x` to `device`, when `x` is an arbitrary
    nested datastructure of dicts and lists containing tensors

    Args:
        x: the tensors
        device: a torch device
        non_blocking (bool): non blocking

    Returns:
        `x` with device changed
    """
    if isinstance(x, torch.Tensor):
        return x.to(device, non_blocking=non_blocking)
    elif isinstance(x, list):
        return [
            send_to_device(xx, device, non_blocking=non_blocking) for xx in x
        ]
    elif isinstance(x, tuple):
        return tuple(
            send_to_device(xx, device, non_blocking=non_blocking) for xx in x)
    elif isinstance(x, dict):
        return {
            k: send_to_device(v, device, non_blocking=non_blocking)
            for k, v in x.items()
        }
    return x


def recursive_state_dict(x: Any) -> Any:
    """
    Recursively call state_dict() on all elements contained in a list / tuple /
    dict so that it can be saved safely via torch.save().

    Args:
        x: any nesting of list / tuple / dict containing state_dict()able
            objects

    Returns:
        the same structure with state dicts
    """
    if hasattr(x, 'state_dict'):
        return x.state_dict()
    if isinstance(x, tuple):
        return tuple(recursive_state_dict(xx) for xx in x)
    if isinstance(x, list):
        return [recursive_state_dict(xx) for xx in x]
    if isinstance(x, dict):
        return {k: recursive_state_dict(v) for k, v in x.items()}
    return x


def load_recursive_state_dict(x: Any, obj: Any) -> None:
    """
    Reload a state dict saved with `recursive_state_dict()`

    Args:
        x: the recursive state dict
        obj: the object that has been recursive_state_dict()ed
    """
    if hasattr(obj, 'load_state_dict'):
        obj.load_state_dict(x)
    if isinstance(x, (tuple, list)):
        for xx, oo in zip(x, obj):
            load_recursive_state_dict(xx, oo)
    if isinstance(x, dict):
        for k in obj.keys():
            load_recursive_state_dict(xx[k], oo[k])


class FrozenModule(nn.Module):
    """
    Wrap a module to eval model, can't be turned back to training mode

    Args:
        m (nn.Module): a module
    """
    def __init__(self, m: nn.Module) -> None:
        super(FrozenModule, self).__init__()
        self.m = freeze(m).eval()

    def train(self, mode: bool = True) -> 'FrozenModule':
        return self

    def __getattr__(self, name):
        print(self, name)
        return getattr(super(FrozenModule, self).__getattr__('m'), name)


class DetachedModule:
    """
    Wrap a module to eval model, can't be turned back to training mode, and
    make it invisible to recursive calls on `nn.Module`s

    Args:
        m (nn.Module): a module
    """
    def __init__(self, m):
        self.m = freeze(m).eval()

    def __call__(self, *args, **kwargs):
        return self.m(*args, **kwargs)

    def __getattr__(self, name):
        return getattr(self.m, name)


@overload
def lerp(a: float, b: float, t: float) -> float:
    ...


@overload
def lerp(a: float, b: float, t: torch.Tensor) -> torch.Tensor:
    ...


@overload
def lerp(a: torch.Tensor, b: torch.Tensor, t: torch.Tensor) -> torch.Tensor:
    ...


@overload
def lerp(a: torch.Tensor, b: torch.Tensor, t: float) -> torch.Tensor:
    ...


def lerp(a, b, t):
    r"""
    Linearly interpolate between `a` and `b` according to `t`.


    :math:`(1 - t)a + tb`

    Args:
        a (number or tensor): a
        b (number or tensor): b
        t (number or tensor): t between 0 and 1

    Returns:
        result between a and b
    """
    return (1 - t) * a + t * b


def ilerp(a: Numeric, b: Numeric, t: Numeric) -> Numeric:
    r"""
    Inverse or lerp. For `t` between `a` and `b`, returns the fraction or `a`
    and `b` in `t`.

    :math:`\frac{t - a}{b - a}`

    Args:
        a (number or tensor): a
        b (number or tensor): b
        t (number or tensor): t between a and b

    Returns:
        result
    """
    return (t - a) / (b - a)


def slerp(z1: torch.Tensor, z2: torch.Tensor, t: float) -> torch.Tensor:
    r"""
    Spherical linear interpolate between `z1` and `z2` according to `t`.

    Args:
        z1 (torch.Tensor): ND tensor, interpolating on last dim
        z2 (torch.Tensor): ND tensor, interpolating on last dim
        t (float): t between 0 and 1

    Returns:
        result between a and b
    """
    z1_l = z1.pow(2).sum(dim=-1, keepdim=True).sqrt()
    z1_n = z1 / z1_l

    z2_l = z2.pow(2).sum(dim=-1, keepdim=True).sqrt()
    z2_n = z2 / z2_l

    dot = torch.sum(z1_n * z2_n, dim=-1).clamp(-1, 1)
    theta_0 = torch.acos(dot)
    theta = t * theta_0

    z3 = z2_n - dot * z1_n
    z3 = z3 / z3.pow(2).sum(dim=-1, keepdim=True).sqrt()

    azimut = lerp(z1_l, z2_l, t)
    return azimut * (z1_n * torch.cos(theta) + z3 * torch.sin(theta))


def as_multiclass_shape(preds, as_probs=False):
    """
    Manipulate an array of logit predictions so that binary prediction is not a
    special case anymore. Outputs `preds` as (Batch size, Num classes (>=2)).

    if preds has one dim, another one is added. If this is binary
    classification, a second column is added as 1 - preds.

    Args:
        preds (tensor): predictions
        as_probs (bool): whether to return the preds as logits or probs

    Returns:
        the predictions reshaped
    """
    assert preds.ndim <= 2

    if preds.ndim == 1:
        preds = preds.unsqueeze(1)
    if preds.shape[1] == 1:
        preds = torch.cat([-preds, preds], dim=1)
        if as_probs:
            preds = torch.sigmoid(preds)
        return preds
    else:
        if as_probs:
            preds = F.softmax(preds, dim=1)
        return preds


class AutoStateDict:
    """
    Inherit this class for automatic :code:`state_dict()` and
    :code:`load_state_dict()` members based on `__dict__`

    Exclusions can be specified via `except_names`
    """
    def __init__(self, except_names: List[str] = []):
        self._except = except_names

    def state_dict(self):
        return {
            key: (val.state_dict() if hasattr(val, 'state_dict') else val)
            for key, val in self.__dict__.items()
            if key not in self._except and key != '_except'
        }

    def load_state_dict(self, state_dict):
        for nm, v in state_dict.items():
            try:
                if hasattr(self.__dict__[nm], 'load_state_dict'):
                    self.__dict__[nm].load_state_dict(v)
                else:
                    self.__dict__[nm] = v
            except KeyError as ke:
                print('no key', ke, 'for', self.__class__.__name__)


def dist_setup(rank):
    """
    initialize a NCCL process group with default port / address. For internal
    use.
    """
    os.environ['MASTER_ADDR'] = 'localhost'
    os.environ['MASTER_PORT'] = '12355'

    # initialize the process group
    dist.init_process_group("nccl",
                            rank=rank,
                            world_size=torch.cuda.device_count())


class _WrapFun:
    def __init__(self, fun, *args, **kwargs):
        self.fun = fun
        self.args = args
        self.kwargs = kwargs

    def __call__(self, rank):
        dist_setup(rank)
        torch.cuda.set_device(rank)
        torch.backends.cudnn.benchmark = True
        return self.fun(*self.args, **self.kwargs, rank=rank)


def parallel_run(fun, *args, n_gpus: int = torch.cuda.device_count(),
                 **kwargs) -> None:
    """
    Starts a function  in parallel for GPU dispatching.

    Args:
        fun (callable): the function to start, with signature
            :code:`fun(rank, world_size, *args, **kwargs)`. :code:`rank` is the
            GPU id tu use on this thread, and :code:`world_size` is the total
            number of GPUs.
        *args: arguments passed to :code:`fun`
        n_gpus (int, optional): number of GPUs to use. Will default to the
            number of available GPUs.
        **kwargs: kw arguments passed to :code:`fun`
    """
    import torch.multiprocessing as mp
    mp.spawn(_WrapFun(fun, *args, **kwargs, world_size=n_gpus),
             nprocs=n_gpus,
             join=True)


def indent(text: str, amount: int = 4) -> str:
    """
    Indent :code:`text` by :code:`amount` spaces.

    Args:
        text (str): some text
        amount (int): an indentation amount

    Returns:
        indented text
    """
    return '\n'.join((' ' * amount + line) for line in text.splitlines())


def experimental(func):
    """
    Decorator that warns about a function being experimental
    """
    msg = (f'{func.__qualname__}() is experimental, '
           'and may change or be deleted soon if not already broken')

    def deprecate_doc(doc):
        if doc is None:
            return f'**Experimental**\n\n.. warning::\n  {msg}\n\n.\n'
        else:
            return ('**Experimental**: ' + dedent(func.__doc__)
                    + f'.. warning::\n {msg}\n\n\n')

    if isfunction(func):
        func.__doc__ = deprecate_doc(func.__doc__)

        @wraps(func)
        def wrapped(*args, **kwargs):
            warnings.warn(msg, FutureWarning, stacklevel=2)
            return func(*args, **kwargs)

        return wrapped
    else:
        cls = func

        def __getstate__(self):
            return super(cls, self).__getstate__()

        def __setstate__(self, state):
            return super(cls, self).__setstate__(state)

        d = {
            '__doc__': deprecate_doc(cls.__doc__),
            '__init__': cls.__init__,
            '__module__': cls.__module__,
            # '__getstate__': __getstate__,
            # '__setstate__': __setstate__
        }
        return type(cls.__name__, (cls, ), d)



from collections import OrderedDict
import torch
import torch.nn as nn

from .condseq import CondSeq
from .conv import ConvBlock, Conv3x3
from .interpolate import InterpolateBilinear2d


class ConvDeconvBlock(nn.Module):
    def __init__(self, in_channels: int, hidden_channels: int,
                 inner: nn.Module) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = in_channels
        self.hidden_channels = hidden_channels
        self.with_skip = False

        self.pre = CondSeq()

        self.downsample = CondSeq(
            OrderedDict([
                ('conv_0',
                 ConvBlock(self.in_channels,
                           self.hidden_channels,
                           4,
                           stride=2)),
            ]))
        self.inner = inner
        assert isinstance(inner.out_channels, int)
        self.upsample = CondSeq(
            OrderedDict([
                ('conv_0',
                 ConvBlock(inner.out_channels,
                           self.in_channels,
                           4,
                           stride=2).to_transposed_conv()),
            ]))

        self.post = CondSeq()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.pre(x)
        out = self.upsample(self.inner(self.downsample(x)))
        if self.with_skip:
            out = torch.cat([out, x], dim=1)
        out = self.post(out)
        return out

    def add_skip(self) -> 'ConvDeconvBlock':
        self.with_skip = True
        if self.with_skip:
            self.out_channels = self.in_channels + self.in_channels
        return self

    def leaky(self) -> 'ConvDeconvBlock':
        for block in [self.pre, self.downsample, self.upsample, self.post]:
            for m in block:
                if isinstance(m, ConvBlock):
                    m.leaky()
        return self


class UBlock(nn.Module):
    downsample: nn.Module

    def __init__(self, in_channels: int, hidden_channels: int,
                 inner: nn.Module) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = in_channels
        self.hidden_channels = hidden_channels

        self.set_encoder_num_layers(2)

        self.downsample = CondSeq(nn.MaxPool2d(2))

        self.inner = inner
        assert isinstance(inner.in_channels, int)
        assert isinstance(inner.out_channels, int)

        self.upsample = CondSeq(
            nn.ConvTranspose2d(inner.out_channels,
                               inner.out_channels // 2,
                               kernel_size=2,
                               stride=2))

        self.set_decoder_num_layers(2)

    def to_bilinear_sampling(self) -> 'UBlock':
        from torchelie.transforms.differentiable import BinomialFilter2d
        self.downsample = BinomialFilter2d(2)
        assert isinstance(self.inner.in_channels, int)
        assert isinstance(self.inner.out_channels, int)
        self.upsample = CondSeq(
            InterpolateBilinear2d(scale_factor=2),
            Conv3x3(self.inner.out_channels, self.inner.out_channels // 2),
            nn.ReLU(True))
        return self

    def forward(self, x_orig: torch.Tensor) -> torch.Tensor:
        x_skip = self.in_conv(x_orig)

        x = self.upsample(self.inner(self.downsample(x_skip)))
        x_skip = torch.cat([x, x_skip], dim=1)

        return self.out_conv(x_skip)

    def set_encoder_num_layers(self, num_layers: int) -> 'UBlock':
        layers = CondSeq()
        for i in range(num_layers):
            layers.add_module(
                f'conv_{i}',
                ConvBlock(
                    self.in_channels if i == 0 else self.hidden_channels,
                    self.hidden_channels, 3))
        self.in_conv = layers
        return self

    def set_decoder_num_layers(self, num_layers: int) -> 'UBlock':
        assert isinstance(self.inner.in_channels, int)
        assert isinstance(self.inner.out_channels, int)

        layers = CondSeq()
        for i in range(num_layers):
            in_ch = (self.inner.out_channels // 2
                     + self.hidden_channels if i == 0 else self.hidden_channels)
            out_ch = (self.in_channels if i == (num_layers
                                                - 1) else self.hidden_channels)

            layers.add_module(f'conv_{i}', ConvBlock(in_ch, out_ch, 3))
        self.out_conv = layers
        return self

    def remove_batchnorm(self) -> 'UBlock':
        for m in self.modules():
            if isinstance(m, ConvBlock):
                m.remove_batchnorm()
        return self

    def leaky(self) -> 'UBlock':
        for m in self.modules():
            if isinstance(m, ConvBlock):
                m.leaky()
        return self

    def set_padding_mode(self, mode: str) -> 'UBlock':
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                m.padding_mode = mode
        return self



import crayons

import torch.nn as nn
from torchelie.utils import experimental


class Debug(nn.Module):
    """
    An pass-through layer that prints some debug info during forward pass.
    It prints its name, the input's shape, mean of channels means, mean,
    mean of channels std, and std.

    Args:
        name (str): this layer's name
    """
    @experimental
    def __init__(self, name):
        super(Debug, self).__init__()
        self.name = name

    def forward(self, x):
        print(crayons.yellow(self.name))
        print(crayons.yellow('----'))
        print('Shape {}'.format(x.shape))
        if x.ndim == 2:
            print("Stats mean {:.2f} var {:.2f}".format(x.mean().item(),
                  x.std().item()))
        if x.ndim == 4:
            print("Stats mean {:.2f} {:.2f} var s{:.2f} {:.2f}".format(
                x.mean(dim=[0, 2, 3]).mean().item(),
                x.mean().item(),
                x.std(dim=[0, 2, 3]).mean().item(),
                x.std().item()))
        print()
        return x


class Dummy(nn.Module):
    """
    A pure pass-through layer
    """
    def forward(self, x):
        return x



import torch.nn as nn


class Lambda(nn.Module):
    """
    Applies a lambda function on forward()

    Args:
        lamb (fn): the lambda function
    """
    def __init__(self, lam):
        super(Lambda, self).__init__()
        self.lam = lam

    def forward(self, x):
        return self.lam(x)


class Reshape(nn.Module):
    """
    Reshape the input volume

    Args:
        *shape (ints): new shape, WITHOUT specifying batch size as first
        dimension, as it will remain unchanged.
    """
    def __init__(self, *shape):
        super(Reshape, self).__init__()
        self.shape = shape

    def forward(self, x):
        return x.view(x.shape[0], *self.shape)



import torch
import torch.nn.functional as F
import torchelie.models as tmodels
from torchelie.metrics.avg import WindowAvg

import torchvision.transforms as TF
from torchvision.datasets import CIFAR10, MNIST, FashionMNIST, SVHN
from torch.utils.data import DataLoader
from torchelie.optim import RAdamW
from torchelie.recipes import TrainAndCall
import torchelie.metrics as tcb


def train(model, loader):
    def train_step(batch):
        x = batch[0]
        x = x.expand(-1, 3, -1, -1)

        x2 = model(x * 2 - 1)
        loss = F.cross_entropy(x2, (x * 255).long())
        loss.backward()
        reconstruction = x2.argmax(dim=1).float() / 255.0
        return {'loss': loss, 'reconstruction': reconstruction}

    def after_train():
        imgs = model.sample(1, 4).expand(-1, 3, -1, -1)
        return {'imgs': imgs}

    opt = RAdamW(model.parameters(), lr=3e-3)
    trainer = TrainAndCall(model,
                           train_step,
                           after_train,
                           dl,
                           test_every=500,
                           visdom_env='pixelcnn')
    trainer.callbacks.add_callbacks([
        tcb.WindowedMetricAvg('loss'),
        tcb.Log('reconstruction', 'reconstruction'),
        tcb.Optimizer(opt, log_lr=True),
        tcb.LRSched(torch.optim.lr_scheduler.ReduceLROnPlateau(opt))
    ])
    trainer.test_loop.callbacks.add_callbacks([
        tcb.Log('imgs', 'imgs'),
    ])

    trainer.to('cuda')
    trainer.run(10)


tfms = TF.Compose([
    TF.Resize(32),
    TF.ToTensor(),
])

dl = DataLoader(FashionMNIST('~/.cache/torch/fashionmnist',
                             transform=tfms,
                             download=True),
                batch_size=32,
                shuffle=True,
                num_workers=4)

model = tmodels.PixelCNN(64, (32, 32), channels=3)
train(model, dl)



import functools

import torch.nn as nn
from torchelie.utils import layer_by_name


class WithSavedActivations(nn.Module):
    """
    Hook :code:`model` in order to get intermediate activations. The
    activations to save can be either specified by module type or layer name.
    """
    def __init__(self, model, types=(nn.Conv2d, nn.Linear), names=None):
        super(WithSavedActivations, self).__init__()
        self.model = model
        self.activations = {}
        self.detach = True
        self.handles = []

        self.set_keep_layers(types, names)

    def set_keep_layers(self, types=(nn.Conv2d, nn.Linear), names=None):
        for h in self.handles:
            h.remove()

        if names is None:
            for name, layer in self.model.named_modules():
                if isinstance(layer, types):
                    h = layer.register_forward_hook(functools.partial(
                        self._save, name))
                    self.handles.append(h)
        else:
            for name in names:
                layer = layer_by_name(self.model, name)
                h = layer.register_forward_hook(functools.partial(
                    self._save, name))
                self.handles.append(h)

    def _save(self, name, module, input, output):
        if self.detach:
            self.activations[name] = output.detach().clone()
        else:
            self.activations[name] = output.clone()

    def forward(self, input, detach: bool):
        """
        Call :code:`self.model(input)`.

        Args:
            input: input to the model
            detach (bool): if True, intermediate activations will be
                :code:`.detach()`d.

        Returns
            model output, a name => activation dict with saved intermediate
            activations.
        """
        self.detach = detach
        self.activations = {}
        out = self.model(input)
        acts = self.activations
        self.activations = {}
        return out, acts



import torch
import pytest
from torchvision.transforms import ToPILImage
from torchelie.transforms import *
import torchelie.transforms.differentiable as dtf


def test_resizenocrop():
    img = ToPILImage()(torch.clamp(torch.randn(3, 32, 16) + 1, min=0, max=1))
    tf = ResizeNoCrop(16)
    assert tf(img).width == 8
    assert tf(img).height == 16


def test_adaptpad():
    img = ToPILImage()(torch.clamp(torch.randn(3, 30, 16) + 1, min=0, max=1))
    tf = AdaptPad((32, 32))
    assert tf(img).width == 32
    assert tf(img).height == 32


def test_multibranch():
    tf = MultiBranch([
        lambda x: x + 1,
        lambda x: x * 3,
        lambda x: x - 1,
    ])
    assert tf(1) == (2, 3, 0)


@pytest.mark.require_opencv
def test_canny():
    img = ToPILImage()(torch.clamp(torch.randn(3, 30, 16) + 1, min=0, max=1))
    tf = Canny()
    tf(img)


def test_resizedcrop():
    img = ToPILImage()(torch.clamp(torch.randn(3, 30, 16) + 1, min=0, max=1))
    tf = ResizedCrop(48)
    tf(img)


def test_diff():
    dtf.roll(torch.randn(3, 16, 16), 3, 3)
    dtf.roll(torch.randn(1, 3, 16, 16), 3, 3)
    dtf.center_crop(torch.randn(3, 16, 16), (4, 4))
    dtf.crop(torch.randn(1, 3, 16, 16))
    dtf.gblur(torch.randn(1, 3, 16, 16))
    dtf.mblur(torch.randn(1, 3, 16, 16))



import logging
from math import log10, sqrt

import numpy as np
import torch
from numpy.fft import fft2, ifft2, ifftshift
from scvi import scvi_logger, settings

"""
Hi-C matrix transformation
"""


def mat2array(mat):
    return mat[np.triu_indices_from(mat, k=0)]


def array2mat(array):
    def _get_square_root(number):
        _number, _tmp = number, 0
        while _number:
            _tmp += 1
            _number -= _tmp
        _number = _tmp
        return _number

    _len = _get_square_root(len(array))

    mat, a = np.zeros((_len, _len)), 0
    for i in range(_len):
        mat[i, i:] = array[a : a + _len - i]
        a += _len - i
    return mat + np.triu(mat, k=1).T


def tensor2mat(tensor: torch.Tensor) -> torch.Tensor:
    _tensor = torch.Tensor(
        np.array(
            [array2mat(tensor[i, :].flatten().cpu()) for i in range(tensor.shape[0])]
        )
    )
    return _tensor.to(tensor.device)


"""
Coarsen the matrix.
"""


def _fftblur(img, sigma):
    h, w = img.shape

    X, Y = np.meshgrid(np.arange(w), np.arange(h))
    X, Y = X - w // 2, Y - h // 2
    Z = np.exp(-0.5 * (X**2 + Y**2) / (sigma**2))
    Z = Z / Z.sum()

    out = ifftshift(ifft2(fft2(img) * fft2(Z)))
    return out


def _fan(sparse_mat, mask):
    N = np.prod(mask.shape)
    num_kept = np.nonzero(mask)[0].shape[0]
    sigma = sqrt(N / (np.pi * num_kept))

    c = _fftblur(sparse_mat, sigma)
    i = _fftblur(mask, sigma)

    mat = np.abs(c / i)
    return mat


def _sparsify(mat, sparse_mat, span):
    h, w = mat.shape
    for i in range(0, h, span):
        for j in range(0, w, span):
            tmp = mat[i : i + span, j : j + span].mean()
            _x, _y = i + span / 2, j + span / 2
            _length = max(abs(_x - _y), 1)
            tmp *= max(log10(_length), 1)
            sparse_mat[i : i + span, j : j + span] += tmp
    return sparse_mat


def coarsen(mat, spans=[5, 10]):
    mask = np.ones(mat.shape)
    sparse_mat = np.zeros(mat.shape)

    spans.sort(reverse=True)
    for span in spans:
        sparse_mat = _sparsify(mat, sparse_mat, span)

    return _fan(sparse_mat, mask)


"""
metric
"""


def hiclip_metric(
    l1_loss: float,
    perceptual_loss: float,
) -> float:
    return np.nanmean([l1_loss, perceptual_loss])


"""
log
"""

_logger = scvi_logger
settings.logging_dir = "./.hiclip/"



import torch.nn as nn
import torchelie.nn as tnn
from torchelie.utils import kaiming
from typing import List
from torchelie.transforms.differentiable import BinomialFilter2d


class PatchDiscriminator(tnn.CondSeq):

    def __init__(self, arch: List[int]) -> None:
        super().__init__()
        layers: List[nn.Module] = [
            tnn.ConvBlock(3, arch[0], kernel_size=4,
                          stride=2).remove_batchnorm().leaky()
        ]

        in_ch = arch[0]
        self.in_channels = in_ch
        for next_ch in arch[1:]:
            layers.append(
                tnn.ConvBlock(in_ch, next_ch, kernel_size=4, stride=2).leaky())
            in_ch = next_ch
        assert isinstance(layers[-1], tnn.ConvBlock)
        layers[-1].conv.stride = (1, 1)

        self.features = tnn.CondSeq(*layers)
        self.classifier = tnn.Conv2d(in_ch, 1, 4)

    def to_equal_lr(self, leak=0.2) -> 'PatchDiscriminator':
        for m in self.modules():
            if isinstance(m, (nn.Linear, nn.Conv2d)):
                kaiming(m, dynamic=True, a=leak)

        return self

    def to_instance_norm(self, affine: bool = True) -> 'PatchDiscriminator':
        """
        Pix2PixHD uses instancenorm rather than batchnorm
        """

        def to_instancenorm(m):
            if isinstance(m, nn.BatchNorm2d):
                return nn.InstanceNorm2d(m.num_features, affine=affine)
            return m

        tnn.utils.edit_model(self, to_instancenorm)

    def to_binomial_downsampling(self) -> 'PatchDiscriminator':
        for m in self.features.modules():
            if isinstance(m, tnn.ConvBlock):
                if m.conv.stride[0] != 2:
                    continue
                tnn.utils.insert_before(m, 'conv', BinomialFilter2d(2), 'pool')
                m.conv.stride = (1, 1)
        return self

    def to_avg_pool(self) -> 'PatchDiscriminator':
        for m in self.features.modules():
            if isinstance(m, tnn.ConvBlock):
                if m.conv.stride[0] != 2:
                    continue
                tnn.utils.insert_before(m, 'conv', nn.AvgPool2d(2), 'pool')
                m.conv.stride = (1, 1)
        return self

    def set_input_specs(self, in_channels: int) -> 'PatchDiscriminator':
        c = self.features[0].conv
        assert isinstance(c, nn.Conv2d)
        self.features[0].conv = kaiming(nn.Conv2d(in_channels,
                                                  c.out_channels,
                                                  4,
                                                  stride=2,
                                                  padding=c.padding,
                                                  bias=c.bias is not None),
                                        a=0.2)
        return self

    def set_kernel_size(self, kernel_size: int) -> 'PatchDiscriminator':

        def change_ks(m):
            if isinstance(m, nn.Conv2d) and m.kernel_size[0] != 1:
                return kaiming(
                    nn.Conv2d(m.in_channels,
                              m.out_channels,
                              kernel_size,
                              m.stride,
                              padding=kernel_size // 2))
            return m

        tnn.utils.edit_model(self.features, change_ks)
        return self


def patch286() -> PatchDiscriminator:
    """
    Patch Discriminator from pix2pix
    """
    return PatchDiscriminator([64, 128, 256, 512, 512, 512])


def patch70() -> PatchDiscriminator:
    """
    Patch Discriminator from pix2pix
    """
    return PatchDiscriminator([64, 128, 256, 512])


def patch34() -> PatchDiscriminator:
    """
    Patch Discriminator from pix2pix
    """
    return PatchDiscriminator([64, 128, 256])


def patch16() -> PatchDiscriminator:
    """
    Patch Discriminator from pix2pix
    """
    return PatchDiscriminator([64, 128])



import torch
from torch.optim import SGD
from torchelie.lr_scheduler import *


def test_curriculum():
    x = torch.randn(3, requires_grad=True)
    opt = SGD([x], 1)
    sched = CurriculumScheduler(opt, [[0, 1, 1], [10, 0, 0]])
    opt.step()
    sched.step()


def test_onecycle():
    x = torch.randn(3, requires_grad=True)
    opt = SGD([x], 1)
    sched = OneCycle(opt, (1e-3, 1e-2), 10)
    opt.step()
    sched.step()



import torch
import torch.nn as nn


class ImageNetInputNorm(nn.Module):
    """
    Normalize images channels as torchvision models expects, in a
    differentiable way
    """
    def __init__(self):
        super(ImageNetInputNorm, self).__init__()
        self.register_buffer(
            'norm_mean',
            torch.tensor([0.485, 0.456, 0.406]).view(1, 3, 1, 1))

        self.register_buffer(
            'norm_std',
            torch.tensor([0.229, 0.224, 0.225]).view(1, 3, 1, 1))

    def forward(self, input):
        return (input - self.norm_mean) / self.norm_std



import collections

import torch
import torch.nn as nn
import torch.nn.functional as F
from .batchnorm import Spade2d
from .condseq import CondSeq
from .maskedconv import MaskedConv2d
from torchelie.utils import kaiming, xavier
from torchelie.utils import experimental
from .layers import ModulatedConv
from .noise import Noise
from .resblock import PreactResBlock
from torchelie.nn.graph import ModuleGraph
from typing import List, Tuple
from .utils import edit_model, make_leaky
from .utils import remove_weight_scale
from .interpolate import InterpolateBilinear2d
from .encdec import ConvDeconvBlock
from .conv import ConvBlock, Conv2d, Conv3x3, Conv1x1


@experimental
def MConvNormReLU(in_ch: int,
                  out_ch: int,
                  ks: int,
                  norm,
                  center: bool = True) -> CondSeq:
    """
    A packed block with Masked Conv-Norm-ReLU

    Args:
        in_ch (int): input channels
        out_ch (int): output channels
        ks (int): kernel size
        norm (ctor): A normalization layer constructor in the form
            :code:`(num_layers) -> norm layer` or None
        center (bool): whether the masked conv has access to the central pixel
            or not

    Returns:
        A packed block with MaskedConv-Norm-ReLU as a CondSeq
    """
    layers: List[Tuple[str, nn.Module]] = []
    layers.append(('conv', MaskedConv2d(in_ch, out_ch, ks, center=center)))

    if norm is not None:
        layers.append(('norm', norm(out_ch)))

    layers.append(('relu', nn.ReLU(True)))
    return CondSeq(collections.OrderedDict(layers))


@experimental
def MConvBNReLU(in_ch: int, out_ch: int, ks: int, center=True) -> CondSeq:
    """
    A packed block with Masked Conv-BN-ReLU

    Args:
        in_ch (int): input channels
        out_ch (int): output channels
        ks (int): kernel size
        center (bool): whether the masked conv has access to the central pixel
            or not

    Returns:
        A packed block with MaskedConv-BN-ReLU as a CondSeq
    """
    return MConvNormReLU(in_ch, out_ch, ks, nn.BatchNorm2d, center=center)


class SpadeResBlock(PreactResBlock):
    """
    A Spade ResBlock from `Semantic Image Synthesis with Spatially-Adaptive
    Normalization`

    https://arxiv.org/abs/1903.07291
    """
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 cond_channels: int,
                 hidden: int,
                 stride: int = 1) -> None:
        super(SpadeResBlock, self).__init__(in_channels=in_channels,
                                            out_channels=out_channels,
                                            stride=stride)

        def to_spade(m: nn.Module) -> nn.Module:
            if isinstance(m, nn.BatchNorm2d):
                return Spade2d(m.num_features, cond_channels, hidden)
            return m

        edit_model(self, to_spade)


class AutoGANGenBlock(nn.Module):
    """
    A block of the generator discovered by AutoGAN.

    Args:
        in_ch (int): number of input channels
        out_ch (int): number of output channels
        skips_ch (list of int): a list with one element per incoming skip
            connection. Each element is the number of channels of the incoming
            skip connection.
        ks (int): kernel size of the convolutions
        mode (str): usampling mode, 'nearest' or 'bilinear'
    """
    def __init__(self,
                 in_ch: int,
                 out_ch: int,
                 skips_ch: List[int],
                 ks: int = 3,
                 mode: str = 'nearest') -> None:
        super(AutoGANGenBlock, self).__init__()
        assert mode in ['nearest', 'bilinear']
        self.mode = mode

        self.preact = CondSeq()

        self.conv1 = ConvBlock(in_ch, out_ch, ks)
        self.conv1.to_preact().remove_batchnorm()
        self.conv1.leaky()

        self.conv2 = ConvBlock(out_ch, out_ch, ks)
        self.conv2.to_preact().remove_batchnorm()
        self.conv2.leaky()

        self.shortcut = None
        if in_ch != out_ch:
            self.shortcut = kaiming(Conv1x1(in_ch, out_ch, 1))
            self.preact.add_module('relu', self.conv1.relu)
            del self.conv1.relu

        self.skip_convs = nn.ModuleList(
            [kaiming(Conv1x1(ch, out_ch)) for ch in skips_ch])

    def forward(
            self,
            x: torch.Tensor,
            skips: List[torch.Tensor] = []
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Forward pass

        Args:
            x (tensor): input tensor
            skips (list of tensor): a tensor per incoming skip connection

        Return:
            output tensor, intermediate value to use in the next block's skip
                connections
        """
        x = F.interpolate(x, scale_factor=2., mode=self.mode)
        x = self.preact(x)
        x_skip = x

        if self.shortcut is not None:
            x_skip = self.shortcut(x_skip)

        x_mid = self.conv1(x)

        x_w_skips = x_mid.clone()
        for conv, skip in zip(self.skip_convs, skips):
            x_w_skips.add_(
                conv(F.interpolate(skip, size=x_mid.shape[-2:],
                                   mode=self.mode)))

        x = self.conv2(x_w_skips)
        return x + x_skip, x_mid


class ResidualDiscrBlock(PreactResBlock):
    """
    A preactivated resblock suited for discriminators: it features leaky relus,
    no batchnorm, and an optional downsampling operator.
    """
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 downsample: bool = False) -> None:
        super().__init__(in_channels, out_channels)
        if downsample:
            self.post.add_module('downsample', nn.AvgPool2d(2))
        self.remove_batchnorm()
        make_leaky(self)

    @experimental
    def to_equal_lr(self) -> 'ResidualDiscrBlock':
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                try:
                    nn.utils.remove_spectral_norm(m)
                except ValueError:
                    pass
                kaiming(m, dynamic=True, a=0.2)

        with torch.no_grad():
            self.branch.conv2.weight_g.zero_()

        return self

    @experimental
    def to_spectral_norm(self) -> 'ResidualDiscrBlock':
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                try:
                    remove_weight_scale(m)
                except ValueError:
                    pass
                nn.utils.spectral_norm(m)

        assert isinstance(self.branch.conv2, nn.Conv2d)
        xavier(self.branch.conv2, a=0.2)
        return self


@experimental
class StyleGAN2Block(nn.Module):
    """
    A Upsample-(ModulatedConv-Noise-LeakyReLU)* block from StyleGAN2

    Args:
        in_ch (int): input channels
        out_ch (int): output channels
        noise_size (int): channels of the conditioning style / noise vector
        upsample (bool): whether to upsample the input or not (default: False)
        n_layers (int): number of conv-noise-relu blocks (default: 2)
        equal_lr (bool): whether to equalize parameters' lr with dynamic weight
            init scaling (weight_scale)

    """
    def __init__(self,
                 in_ch: int,
                 out_ch: int,
                 noise_size: int,
                 upsample: bool = False,
                 n_layers: int = 2,
                 equal_lr: bool = True):
        super().__init__()
        self.upsample_mode = 'bilinear'
        self.equal_lr = equal_lr
        dyn = equal_lr
        self.upsample = upsample
        inside = ModuleGraph(outputs=f'in_{n_layers}')

        for i in range(n_layers):
            conv = ModulatedConv(in_ch,
                                 noise_size,
                                 out_ch,
                                 kernel_size=3,
                                 padding=1,
                                 bias=True)
            kaiming(conv, dynamic=dyn, a=0.2)
            inside.add_operation(inputs=[f'in_{i}', 'w'],
                                 outputs=[f'conv_{i}'],
                                 name=f'conv_{i}',
                                 operation=conv)

            noise = Noise(out_ch, inplace=True, bias=False)
            inside.add_operation(inputs=[f'conv_{i}', f'noise_{i}'],
                                 operation=noise,
                                 name=f'plus_noise_{i}',
                                 outputs=[f'plus_noise_{i}'])

            inside.add_operation(inputs=[f'plus_noise_{i}'],
                                 operation=nn.LeakyReLU(0.2, True),
                                 outputs=[f'in_{i+1}'],
                                 name=f'relu_{i+1}')

            in_ch = out_ch

        if dyn:
            for m in inside:
                if isinstance(m, ModulatedConv):
                    xavier(m.make_s, dynamic=True)

        self.inside = inside
        self.to_rgb = xavier(ModulatedConv(out_ch,
                                           noise_size,
                                           3,
                                           kernel_size=1,
                                           padding=0,
                                           bias=True,
                                           demodulate=False),
                             dynamic=dyn,
                             a=0.2)

    def n_noise(self) -> int:
        return len(self.inside) // 3

    def extra_repr(self):
        return (f"upsample={self.upsample} n_layers={len(self.inside) // 3} "
                f"equal_lr={self.equal_lr}")

    def forward(self, maps, w, rgb, noises=None):
        if rgb is None:
            rgb = torch.zeros(maps.shape[0],
                              3,
                              *maps.shape[2:],
                              device=maps.device)

        if self.upsample:
            rgb = nn.functional.interpolate(rgb,
                                            scale_factor=2,
                                            mode=self.upsample_mode,
                                            align_corners=False,
                                            recompute_scale_factor=True)
            maps = nn.functional.interpolate(maps,
                                             scale_factor=2,
                                             mode=self.upsample_mode,
                                             align_corners=False,
                                             recompute_scale_factor=True)
        maps = self.inside(
            in_0=maps,
            w=w,
            **({f'noise_{i}': None
                for i in range(len(self.inside) // 3)}))
        return rgb + self.to_rgb(maps, w), maps



import math
import torchelie.utils as tu
from typing import List, Tuple


class CurriculumScheduler:
    """
    Allow to pre-specify learning rate and momentum changes

    Args:
        optimizer (torch.optim.Optimizer): the optimizer to schedule. Currently
            works only with SGD
        schedule (list): a schedule. It's a list of keypoints where each
            element is a 3-tuple like (iteration number, lr, mom). Values are
            interpolated linearly between neighboring keypoints
        last_iter (int): starting iteration
    """
    def __init__(self,
                 optimizer,
                 schedule: List[Tuple[float, float, float]],
                 last_iter: int = -1):
        self.optimizer = optimizer
        self.schedule = schedule
        self.last_iter = last_iter

    def state_dict(self):
        return {'last_iter': self.last_iter}

    def load_state_dict(self, state):
        self.last_iter = state['last_iter']

    def step(self, *unused) -> None:
        """
        Step the scheduler to another iteration
        """
        self.last_iter += 1
        the_lr = self.schedule[-1][1]
        the_mom = self.schedule[-1][2]
        for lo, hi in zip(self.schedule[:-1], self.schedule[1:]):
            limit_lo, lr_lo, mom_lo = lo
            lim_hi, lr_hi, mom_hi = hi

            if limit_lo <= self.last_iter < lim_hi:
                t = tu.ilerp(limit_lo, lim_hi, self.last_iter)
                the_lr = tu.lerp(lr_lo, lr_hi, t)
                the_mom = tu.lerp(mom_lo, mom_hi, t)

        for group in self.optimizer.param_groups:
            group['lr'] = the_lr
            if 'momentum' in group:
                group['momentum'] = the_mom
            elif 'betas' in group:
                group['betas'] = (the_mom, group['betas'][1])

    def __repr__(self):
        return "CurriculumScheduler({})".format(self.schedule)


class OneCycle(CurriculumScheduler):
    """
    Implements 1cycle policy.

    Goes from:
    - `lr[0]` to `lr[1]` during `num_iters // 3` iterations
    - `lr[1]` to `lr[0]` during another `num_iters // 3` iterations
    - `lr[0]` to `lr[0] // 10` during another `num_iters // 3` iterations

    Args:
        opt (Optimizer): the optimizer on which to modulate lr
        lr (2-tuple): lr range
        num_iters (int): total number of iterations
        mom (2-tuple): momentum range
        last_iter (int): last_iteration index
    """
    def __init__(self,
                 opt,
                 lr: Tuple[float, float],
                 num_iters: int,
                 mom: Tuple[float, float] = (0.95, 0.85),
                 log: bool = False,
                 last_iter: int = -1):
        self.log = log

        if log:
            lr = math.log(lr[0]), math.log(lr[1])
            mom = math.log(mom[0]), math.log(mom[1])

        third = num_iters // 3
        super(OneCycle, self).__init__(
            opt, [(0, lr[0], mom[0]), (third, lr[1], mom[1]),
                  (2 * third, lr[0], mom[0]),
                  (num_iters, lr[0] / (lr[1] / lr[0] * 10000), mom[0])],
            last_iter=last_iter)

    def step(self, *unused):
        super(OneCycle, self).step()

        if not self.log:
            return

        for group in self.optimizer.param_groups:
            group['lr'] = math.exp(group['lr'])
            if 'momentum' in group:
                group['momentum'] = math.exp(group['momentum'])
            elif 'betas' in group:
                group['betas'] = (math.exp(group['betas'][0]),
                                  group['betas'][1])

    def __repr__(self):
        return 'OneCycle({})'.format(self.schedule)



from torchelie.datasets import *


class StupidDataset:
    def __init__(self):
        self.classes = [0, 1]
        self.imgs = [(i, 0 if i < 5 else 1) for i in range(10)]
        self.samples = self.imgs

    def __len__(self):
        return 10

    def __getitem__(self, i):
        return torch.FloatTensor([i]), torch.LongTensor([self.imgs[i][1]])


def test_colored():
    cc = ColoredColumns(64, 64)
    cr = ColoredRows(64, 64)

    cc[0]
    cr[0]


def test_paired():
    ds = StupidDataset()
    ds2 = StupidDataset()

    pd = PairedDataset(ds, ds2)
    assert len(pd) == 100
    assert pd[0] == list(zip(ds[0], ds2[0]))


def test_cat():
    ds = StupidDataset()
    ds2 = StupidDataset()

    cated = HorizontalConcatDataset([ds, ds2])
    assert len(cated) == len(ds) * 2
    print([cated[i][1] for i in range(20)])

    for i in range(len(ds)):
        assert cated[i][1] == (0 if i < 5 else 1)
        assert cated[len(ds) + i][1] == (2 if i < 5 else 3)


def test_mixup():
    ds = StupidDataset()
    md = MixUpDataset(ds)
    md[0]


def test_cached():
    ds = StupidDataset()
    md = CachedDataset(ds)
    md[0]


def test_withindex():
    ds = StupidDataset()
    md = WithIndexDataset(ds)
    md[0]



import torch
from collections import defaultdict, OrderedDict
import torchelie.utils as tu


class CallbacksRunner:
    def __init__(self):
        self.cbs = [[], [], []]
        self.reset()

    def reset(self):
        self.state = {'metrics': {}}

    def __call__(self, name, *args, **kwargs):
        for cb in self.callbacks():
            if hasattr(cb, name):
                getattr(cb, name)(self.state, *args, **kwargs)

    def callbacks(self):
        for cbs in self.cbs:
            for cb in cbs:
                yield cb

    def named_callbacks(self):
        for step, cbs in zip(['prologue', 'middle', 'epilogue'], self.cbs):
            counts = defaultdict(int)
            for cb in cbs:
                nm = cb.__class__.__name__
                cnt = counts[nm]
                counts[nm] += 1
                yield '_'.join([nm, step, str(cnt)]), cb

    def state_dict(self):
        serial_cb = {}
        for nm, cb in self.named_callbacks():
            if hasattr(cb, 'state_dict'):
                serial_cb[nm] = cb.state_dict()
        return {'state': self.state, 'callbacks': serial_cb}

    def load_state_dict(self, dicc):
        self.state = dicc['state']

        for nm, cb in self.named_callbacks():
            if hasattr(cb, 'load_state_dict') and nm in dicc['callbacks']:
                cb.load_state_dict(dicc['callbacks'][nm])

    def update_state(self, state_additions):
        self.state.update(state_additions)

    def add_prologue(self, cb):
        self.cbs[0].append(cb)

    def add_callback(self, cb):
        self.cbs[1].append(cb)

    def add_epilogue(self, cb):
        self.cbs[2].append(cb)

    def add_prologues(self, cbs):
        for cb in cbs:
            self.add_prologue(cb)

    def add_callbacks(self, cbs):
        for cb in cbs:
            self.add_callback(cb)

    def add_epilogues(self, cbs):
        for cb in cbs:
            self.add_epilogue(cb)

    def __repr__(self):
        return "Prologue:\n{}\nCallbacks:\n{}\nEpilogue:\n{}".format(
            "\n".join(["  " + line for line in repr(self.cbs[0]).split("\n")]),
            "\n".join(["  " + line for line in repr(self.cbs[1]).split("\n")]),
            "\n".join(["  " + line for line in repr(self.cbs[2]).split("\n")]),
        )


class RecipeBase:
    def __init__(self):
        self._modules = set()
        self._savable = set()
        self.device = 'cpu'

    def _check_init(self):
        if '_modules' not in self.__dict__:
            raise AttributeError('You forgot to call ModulesAware.__init__()')

    def register(self, name, value):
        """
        Register an object into the recipe as a member. Calling
        :code:`recipe.register('foo', bar)` registers bar, and makes it usable
        through :code:`recipe.foo`.

        Args:
            name (str): member's name
            value: the object to register
        """
        self._check_init()
        self._modules.discard(name)
        self._savable.discard(name)

        if isinstance(value, torch.nn.Module):
            self._modules.add(name)
        else:
            self._savable.add(name)

        self.__dict__[name] = value

    def state_dict(self):
        """
        Returns:
            A state dict
        """
        sd = OrderedDict()
        for nm in self._modules:
            mod = self.__dict__[nm]
            if isinstance(mod, (torch.nn.parallel.DistributedDataParallel,
                                torch.nn.parallel.DataParallel)):
                sd[nm] = mod.module.state_dict()
            else:
                sd[nm] = mod.state_dict()

        for nm in self._savable:
            val = self.__dict__[nm]
            if hasattr(val, 'state_dict'):
                sd[nm] = val.state_dict()
            else:
                sd[nm] = val
        return sd

    def load_state_dict(self, state_dict):
        """
        Restore a recipe
        """
        for key, state in state_dict.items():
            val = self.__dict__[key]
            if hasattr(val, 'load_state_dict'):
                if isinstance(val, torch.nn.Module):
                    if isinstance(val,
                                  (torch.nn.parallel.DistributedDataParallel,
                                   torch.nn.parallel.DataParallel)):
                        print(val.module.load_state_dict(state, strict=False))
                    else:
                        print(val.load_state_dict(state, strict=False))
                else:
                    val.load_state_dict(state)
            else:
                self.__dict__[key] = val
        return self

    def modules(self):
        """
        Iterate over all nn.Modules registered in the recipe
        """
        for m in self._modules:
            yield self.__dict__[m]

    def to(self, device):
        """
        Move a recipe and all its movable registered objects to a device

        Args:
            device: a torch device
        """
        self._check_init()
        self.device = device

        for m in self.modules():
            m.to(self.device)

        for nm in self._savable:
            m = self.__dict__[nm]
            if hasattr(m, 'to'):
                self.__dict__[nm] = m.to(self.device)

        return self

    def cuda(self):
        """
        Move a recipe and all its movable registered objects to cuda
        """
        return self.to('cuda')

    def cpu(self):
        """
        Move a recipe and all its movable registered objects to cpu
        """
        return self.to('cpu')


class Recipe(RecipeBase):
    """
    Basic recipe that iterates mutiple epochs over a dataset. That loop is
    instrumented through several configurable callbacks. Callbacks can handle
    events before and after each batch, before and after each epoch. Each batch
    is treated with a user supplied function that manipulates it and returns a
    dict that returns a state usable by the callbacks.

    A recipe can be saved by calling its :code:`state_dict()` member. All its
    hyper parameters and state will be saved so that it can restart, but
    requires the exact same setting of callbacks.

    You can register multiple objects to a Recipe with :code:`register()`. If
    it has a :code:`state_dict()` member, its state will be saved into the
    recipe's when calling the recipe's :code:`state_dict()` function. If it
    has a member :code:`to()`, moving the recipe to another device will also
    move those objects.

    Args:
        call_fun (Callable): A function that takes a batch as an argument and
            returns a dict of value to feed the state
        loader (Iterable): any iterable (most likely a DataLoader)
    """
    def __init__(self, call_fun, loader):
        super(Recipe, self).__init__()
        self.call_fun = call_fun
        self.loader = loader

        self.callbacks = CallbacksRunner()
        self.register('callbacks', self.callbacks)

    def run(self, epochs):
        """
        Run the recipe for :code:`epochs` epochs.

        Args:
            epochs (int): number of epochs

        Returns:
            The state
        """
        self.to(self.device)
        for epoch in range(epochs):
            self.callbacks('on_epoch_start')
            for batch in self.loader:
                self.callbacks.update_state({'batch': batch})
                batch = tu.send_to_device(batch,
                                          self.device,
                                          non_blocking=True)
                self.callbacks.update_state({'batch_gpu': batch})

                self.callbacks('on_batch_start')
                out = self.call_fun(batch)
                out = tu.send_to_device(out, 'cpu', non_blocking=False)
                self.callbacks.update_state(out)
                self.callbacks('on_batch_end')
            self.callbacks('on_epoch_end')
        return self.callbacks.state



import torch.nn as nn
from typing import Tuple, Optional, Union, cast
from .utils import remove_batchnorm, insert_after, insert_before
from .condseq import CondSeq
from torchelie.utils import kaiming, experimental
from .interpolate import InterpolateBilinear2d


def Conv2d(in_ch, out_ch, ks, stride=1, bias=True) -> nn.Conv2d:
    """
    A Conv2d with 'same' padding
    """
    return nn.Conv2d(in_ch,
                     out_ch,
                     ks,
                     padding=(ks - 1) // 2,
                     stride=stride,
                     bias=bias)


def Conv3x3(in_ch: int,
            out_ch: int,
            stride: int = 1,
            bias: bool = True) -> nn.Conv2d:
    """
    A 3x3 Conv2d with 'same' padding
    """
    return Conv2d(in_ch, out_ch, 3, stride=stride, bias=bias)


def Conv1x1(in_ch: int,
            out_ch: int,
            stride: int = 1,
            bias: bool = True) -> nn.Conv2d:
    """
    A 1x1 Conv2d
    """
    return Conv2d(in_ch, out_ch, 1, stride=stride, bias=bias)


class ConvBlock(CondSeq):
    """
    A packed block with Conv-BatchNorm-ReLU and various operations to alter it.

    Args:
        in_channels (int): input channels
        out_channels (int): output channels
        kernel_size (int): kernel size
        stride (int): stride of the conv

    Returns:
        A packed block with Conv-Norm-ReLU as a CondSeq
    """
    conv: Union[nn.Conv2d, nn.ConvTranspose2d]
    norm: Optional[nn.Module]
    relu: Optional[nn.Module]

    out_channels: int
    in_channels: int
    kernel_size: Tuple[int, int]
    stride: Tuple[int, int]

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: int,
                 stride: int = 1) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = (kernel_size, kernel_size)
        self.stride = (stride, stride)
        self.reset()

    def reset(self) -> None:
        """
        Recreate the block as a simple conv-BatchNorm-ReLU
        """
        self.add_module(
            'conv',
            kaiming(
                Conv2d(self.in_channels,
                       self.out_channels,
                       self.kernel_size[0],
                       stride=self.stride,
                       bias=False)))
        self.add_module('norm', nn.BatchNorm2d(self.out_channels))
        self.add_module('relu', nn.ReLU(inplace=True))

    def to_input_specs(self, in_channels: int) -> 'ConvBlock':
        """
        Recreate a convolution with :code:`in_channels` input channels
        """
        self.in_channels = in_channels

        c = self.conv
        self.conv = kaiming(
            Conv2d(in_channels,
                   c.out_channels,
                   c.kernel_size[0],
                   stride=c.stride,
                   bias=c.bias))
        return self

    @experimental
    def to_transposed_conv(self) -> 'ConvBlock':
        """
        Transform the convolution into a hopefully equivalent transposed
        convolution
        """
        c = self.conv
        assert c.kernel_size[0] in [
            3, 4
        ], ('ConvBlock.to_transposed_conv()'
            ' not supported with kernel_size other than 2 or 3 (please add the'
            ' support!')
        if c.kernel_size[0] == 4:
            self.conv = kaiming(
                nn.ConvTranspose2d(c.in_channels,
                                   c.out_channels,
                                   cast(Tuple[int, int], c.kernel_size),
                                   stride=cast(Tuple[int, int], c.stride),
                                   padding=cast(Tuple[int, int], c.padding),
                                   bias=c.bias is not None,
                                   padding_mode=c.padding_mode))
        else:
            self.conv = kaiming(
                nn.ConvTranspose2d(c.in_channels,
                                   c.out_channels,
                                   cast(Tuple[int, int], c.kernel_size),
                                   stride=cast(Tuple[int, int], c.stride),
                                   padding=cast(Tuple[int, int], c.padding),
                                   output_padding=cast(Tuple[int, int],
                                                       c.padding),
                                   bias=c.bias is not None,
                                   padding_mode=c.padding_mode))
        return self

    def remove_batchnorm(self) -> 'ConvBlock':
        """
        Remove the BatchNorm, restores the bias term in conv.

        Returns:
            self
        """
        remove_batchnorm(self)
        if hasattr(self, 'norm'):
            del self.norm
        self.norm = None
        return self

    def add_upsampling(self) -> 'ConvBlock':
        """
        Add a bilinear upsampling layer before the conv that doubles the
        spatial size
        """
        insert_before(self, 'conv', InterpolateBilinear2d(scale_factor=2),
                      'upsample')
        return self

    def restore_batchnorm(self) -> 'ConvBlock':
        """
        Restore BatchNorm if deleted
        """
        if self.norm is not None:
            return self
        insert_after(self, 'conv', nn.BatchNorm2d(self.out_channels), 'norm')
        return self

    def no_bias(self) -> 'ConvBlock':
        """
        Remove the bias term.

        Returns:
            self
        """
        if hasattr(self.norm, 'bias') and self.norm is not None:
            self.norm.bias = None  # type: ignore
        self.conv.bias = None
        return self

    def leaky(self, leak: float = 0.2) -> 'ConvBlock':
        """
        Change the ReLU to a LeakyReLU, also rescaling the weights in the conv
        to preserve the variance.

        Returns:
            self
        """
        if not hasattr(self, 'relu') or self.relu is None:
            return self

        new_gain = nn.init.calculate_gain('leaky_relu', param=leak)
        if isinstance(self.relu, nn.LeakyReLU):
            old_gain = nn.init.calculate_gain('leaky_relu',
                                              param=self.relu.negative_slope)
        else:
            old_gain = nn.init.calculate_gain('relu')

        self.relu = nn.LeakyReLU(leak, inplace=True)
        self.conv.weight.data *= new_gain / old_gain
        return self

    def no_relu(self) -> 'ConvBlock':
        """
        Remove the ReLU
        """
        del self.relu
        self.relu = None
        return self

    def to_preact(self) -> 'ConvBlock':
        """
        Place the normalization and ReLU before the convolution.
        """
        self.reset()
        c = self.conv
        del self.conv
        self.add_module('conv', c)
        return self



import torch
import torch.nn as nn
import torchelie.nn as tnn
from torchelie.utils import kaiming
from typing import cast
from .classifier import ClassificationHead


class VGG(nn.Module):
    def __init__(self, arch: list, num_classes: int) -> None:
        super().__init__()
        self.arch = arch
        in_ch = 3
        self.in_channels = in_ch

        feats = tnn.CondSeq()
        block_num = 1
        conv_num = 1
        for layer in arch:
            if layer == 'M':
                feats.add_module(f'pool_{block_num}', nn.MaxPool2d(2, 2))
                block_num += 1
                conv_num = 1
            else:
                ch = cast(int, layer)
                feats.add_module(
                    f'conv_{block_num}_{conv_num}',
                    tnn.ConvBlock(in_ch, ch, 3).remove_batchnorm())
                in_ch = ch
                conv_num += 1
        self.out_channels = ch

        self.features = feats
        self.classifier = ClassificationHead(self.out_channels, num_classes)
        self.classifier.to_vgg_style(4096)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.classifier(self.features(x))

    def add_bn(self) -> 'VGG':
        for m in self.features:
            if isinstance(m, tnn.ConvBlock):
                m.restore_bn()
        return self

    def set_input_specs(self, in_channels: int) -> 'VGG':
        c1 = self.features.conv_1_1
        assert isinstance(c1, tnn.ConvBlock)
        c1.conv = kaiming(tnn.Conv3x3(in_channels, c1.conv.out_channels, 3))
        return self


def vgg11(num_classes: int) -> 'VGG':
    return VGG(
        [64, 'M', 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
        num_classes)


def vgg13(num_classes: int) -> 'VGG':
    return VGG([
        64, 64, 'M', 128, 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'
    ], num_classes)


def vgg16(num_classes: int) -> 'VGG':
    return VGG([
        64, 64, 'M', 128, 128, 'M', 256, 256, 256, 'M', 512, 512, 512, 'M',
        512, 512, 512, 'M'
    ], num_classes)


def vgg19(num_classes: int) -> 'VGG':
    return VGG([
        64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M', 512, 512, 512,
        512, 'M', 512, 512, 512, 512, 'M'
    ], num_classes)


def vgg11_bn(num_classes: int) -> 'VGG':
    return vgg11(num_classes).add_bn()


def vgg13_bn(num_classes: int) -> 'VGG':
    return vgg13(num_classes).add_bn()


def vgg16_bn(num_classes: int) -> 'VGG':
    return vgg16(num_classes).add_bn()


def vgg19_bn(num_classes: int) -> 'VGG':
    return vgg19(num_classes).add_bn()



import torch
from PIL import Image
from io import BytesIO
import struct
from typing import List, Tuple, Any


class MS1M:
    def __init__(self, rec_file: str, idx_file: str, transform=None) -> None:
        self.rec_file = rec_file
        self.idx_file = idx_file

        offsets = self.read_idx(idx_file)
        max_id = int(self.read_metadata(offsets[-1])[1][0]) - 1
        self.transform = transform
        try:
            self.samples = torch.load('ms1m_cache.pth')
        except:
            self.samples = [(str(off), int(self.read_metadata(off)[1]))
                            for off, i in zip(offsets, range(max_id))]
            torch.save(self.samples, 'ms1m_cache.pth')
        self.classes = [str(i) for i in range(85742)]
        self.class_to_idx = {k: i for i, k in enumerate(self.classes)}
        self.imgs = self.samples

    def __len__(self) -> int:
        return len(self.samples)

    @staticmethod
    def read_idx(idx_file: str) -> List[int]:
        indices = []
        with open(idx_file, 'r') as f:
            for line in f.readlines():
                n, offset = line.strip().split('\t')
                indices.append(int(offset))
        return indices

    def read_metadata(self, offset: int) -> Tuple[bytes, Any]:
        with open(self.rec_file, 'rb') as rec_handle:
            rec_handle.seek(offset)
            magic, lrec = struct.unpack('<II', rec_handle.read(8))
            assert magic == 0xced7230a, f"{hex(magic)}, {hex(lrec)}"
            cflag = lrec >> 29
            assert cflag == 0
            length = lrec & ~(3 << 29)
            header_sz = struct.calcsize('IfQQ')
            flag, label, id1, id2 = struct.unpack('IfQQ',
                                                  rec_handle.read(header_sz))
            if flag > 0:
                label = struct.unpack('f' * flag, rec_handle.read(4 * flag))
                header_sz -= 4 * flag
            img_bytes = rec_handle.read(length - header_sz)
            return img_bytes, label

    def __getitem__(self, i: int) -> Tuple[Any, int]:
        offset, label = self.samples[i]
        img_bytes, _ = self.read_metadata(int(offset))
        assert img_bytes[:3] == b'\xff\xd8\xff'

        with BytesIO(img_bytes) as dat:
            img = Image.open(dat).convert('RGB')

        if self.transform is not None:
            img = self.transform(img)

        return img, int(label)

    def __repr__(self) -> str:
        return (f"MS1M Dataset:\n"
                f"  n_samples: {len(self.samples)}\n"
                f"  n_classes: {len(self.classes)}\n")



import torch
import torch.nn as nn
import torchelie.utils as tu

from typing import Optional


class AdaIN2d(nn.Module):
    """
    Adaptive InstanceNormalization from *Arbitrary Style Transfer in Real-time
    with Adaptive Instance Normalization* (Huang et al, 2017)

    Args:
        channels (int): number of input channels
        cond_channels (int): number of conditioning channels from which bias
            and scale will be derived
    """
    weight: torch.Tensor
    bias: torch.Tensor

    def __init__(self, channels: int, cond_channels: int) -> None:
        super(AdaIN2d, self).__init__()
        self.make_weight = nn.Linear(cond_channels, channels)
        self.make_bias = nn.Linear(cond_channels, channels)
        self.register_buffer('weight', torch.zeros(0))
        self.register_buffer('bias', torch.zeros(0))

    def forward(self,
                x: torch.Tensor,
                z: Optional[torch.Tensor] = None) -> torch.Tensor:
        """
        Forward pass

        Args:
            x (4D tensor): input tensor
            z (2D tensor, optional): conditioning vector. If not present,
                :code:`condition(z)` must be called first

        Returns:
            x, renormalized
        """
        if z is not None:
            self.condition(z)

        m = x.mean(dim=(2, 3), keepdim=True)
        s = torch.sqrt(x.var(dim=(2, 3), keepdim=True) + 1e-8)

        weight = self.weight / (s + 1e-5)
        bias = -m * weight + self.bias
        out = weight * x + bias
        return out

    def condition(self, z: torch.Tensor) -> None:
        """
        Conditions the layer before the forward pass if z will not be present
        when calling forward

        Args:
            z (2D tensor, optional): conditioning vector
        """
        self.weight = self.make_weight(z)[:, :, None, None] + 1
        self.bias = self.make_bias(z)[:, :, None, None]


class FiLM2d(nn.Module):
    """
    Feature-wise Linear Modulation from
    https://distill.pub/2018/feature-wise-transformations/
    The difference with AdaIN is that FiLM does not uses the input's mean and
    std in its calculations

    Args:
        channels (int): number of input channels
        cond_channels (int): number of conditioning channels from which bias
            and scale will be derived
    """
    weight: Optional[torch.Tensor]
    bias: Optional[torch.Tensor]

    def __init__(self, channels: int, cond_channels: int) -> None:
        super(FiLM2d, self).__init__()
        self.make_weight = nn.Linear(cond_channels, channels)
        tu.normal_init(self.make_weight, 0.01)
        self.make_bias = nn.Linear(cond_channels, channels)
        tu.normal_init(self.make_bias, 0.01)

        self.weight = None
        self.bias = None

    def forward(self, x, z: Optional[torch.Tensor] = None) -> torch.Tensor:
        """
        Forward pass

        Args:
            x (4D tensor): input tensor
            z (2D tensor, optional): conditioning vector. If not present,
                :code:`condition(z)` must be called first

        Returns:
            x, conditioned
        """
        if z is not None:
            self.condition(z)

        w = self.weight
        b = self.bias
        assert w is not None and b is not None
        return w * x + b

    def condition(self, z: torch.Tensor) -> None:
        """
        Conditions the layer before the forward pass if z will not be present
        when calling forward

        Args:
            z (2D tensor, optional): conditioning vector
        """
        self.weight = self.make_weight(z)[:, :, None, None].add_(1)
        self.bias = self.make_bias(z)[:, :, None, None]



from typing import Optional
import torch
import torch.nn as nn
import torch.nn.functional as F


def ortho(w: torch.Tensor) -> torch.Tensor:
    r"""
    Returns the orthogonal loss for weight matrix `m`, from Big GAN.

    https://arxiv.org/abs/1809.11096

    :math:`R_{\beta}(W)= ||W^T W  \odot (1 - I)||_F^2`
    """
    cosine = torch.einsum('ij,ji->ij', w, w)
    no_diag = (1 - torch.eye(w.shape[0], device=w.device))
    return (cosine * no_diag).pow(2).sum(dim=1).mean()


def total_variation(i: torch.Tensor) -> torch.Tensor:
    """
    Returns the total variation loss for batch of images `i`
    """
    v = F.l1_loss(i[:, :, 1:, :], i[:, :, :-1, :])
    h = F.l1_loss(i[:, :, :, 1:], i[:, :, :, :-1])
    return v + h


def focal_loss(input: torch.Tensor,
               target: torch.Tensor,
               gamma: float = 0,
               weight: Optional[torch.Tensor] = None) -> torch.Tensor:
    r"""
    Returns the focal loss between `target` and `input`

    :math:`\text{FL}(p_t)=-(1-p_t)^\gamma\log(p_t)`
    """
    if input.shape[1] == 1:
        logp = nn.functional.binary_cross_entropy_with_logits(input, target)
    else:
        logp = nn.functional.cross_entropy(input, target, weight=weight)
    p = torch.exp(-logp)
    loss = (1 - p)**gamma * logp
    return loss.mean()


def continuous_cross_entropy(pred: torch.Tensor,
                             soft_targets: torch.Tensor,
                             weights: Optional[torch.Tensor] = None,
                             reduction: str = 'mean') -> torch.Tensor:
    r"""
    Compute the cross entropy between the logits `pred` and a normalized
    distribution `soft_targets`. If `soft_targets` is a one-hot vector, this is
    equivalent to `nn.functional.cross_entropy` with a label
    """
    if weights is None:
        ce = torch.sum(-soft_targets * F.log_softmax(pred, 1), 1)
    else:
        ce = torch.sum(-weights * soft_targets * F.log_softmax(pred, 1), 1)

    if reduction == 'mean':
        return ce.mean()
    if reduction == 'sum':
        return ce.sum()
    if reduction == 'none':
        return ce
    assert False, f'{reduction} not a valid reduction method'
